/***********************************************************/
/*                                                         */
/* D3DStructures.h                                         */
/*                                                         */
/* created 20.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef D3DSTRUCTURES_H
#define D3DSTRUCTURES_H

#include <d3d8.h>

#include "MathStructures.h"
#include "DynamicArray.h"

// www.infinitecode.com 3d library
namespace I3D
{
	using IMATH::Vector;
	using IMATH::Vector2D;

#ifndef RELEASE
	#define RELEASE(x) if(x){(x)->Release();}
#endif

#define I3D_FVFVERTEX (D3DFVF_XYZ | D3DFVF_NORMAL|D3DFVF_TEX1)
#define I3D_FVFVERTEXTEX2 (D3DFVF_XYZ | D3DFVF_NORMAL|D3DFVF_TEX2)

class Vertex
{
public:
	union
	{
		struct
		{
			float x, y, z, nx, ny, nz, tu, tv;
		};
		struct
		{
			Vector v, n;
			Vector2D t;
		};
		float Data[8];
	};

	Vertex() {}

	Vertex(Vector Vec, Vector Norm, Vector2D Tex)
	{
		x = Vec.x;
		y = Vec.y;
		z = Vec.z;
		nx = Norm.x;
		ny = Norm.y;
		nz = Norm.z;
		tu = Tex.x;
		tv = Tex.y;
	}

	void Set(Vector Vec, Vector Norm, Vector2D Tex)
	{
		x = Vec.x;
		y = Vec.y;
		z = Vec.z;
		nx = Norm.x;
		ny = Norm.y;
		nz = Norm.z;
		tu = Tex.x;
		tv = Tex.y;
	}
};

class VertexTex2
{
public:
	union
	{
		struct
		{
			float x, y, z, nx, ny, nz, tu0, tv0, tu1, tv1;
		};
		struct
		{
			Vector v, n;
			Vector2D t0, t1;
		};
		float Data[10];
	};

	VertexTex2() {}

	VertexTex2(Vector Vec, Vector Norm, Vector2D Tex0, Vector2D Tex1)
	{
		x = Vec.x;
		y = Vec.y;
		z = Vec.z;
		nx = Norm.x;
		ny = Norm.y;
		nz = Norm.z;
		tu0 = Tex0.x;
		tv0 = Tex0.y;
		tu1 = Tex1.x;
		tv1 = Tex1.y;
	}

	void Set(Vector Vec, Vector Norm, Vector2D Tex0, Vector2D Tex1)
	{
		x = Vec.x;
		y = Vec.y;
		z = Vec.z;
		nx = Norm.x;
		ny = Norm.y;
		nz = Norm.z;
		tu0 = Tex0.x;
		tv0 = Tex0.y;
		tu1 = Tex1.x;
		tv1 = Tex1.y;
	}
};

class Matrix: public D3DMATRIX
{
public:
	Vector operator * (Vector V) const
	{
		Vector r;
		// apply
		r.x = V.x * m[0][0] + V.y * m[1][0] + V.z * m[2][0] + m[3][0];
		r.y = V.x * m[0][1] + V.y * m[1][1] + V.z * m[2][1] + m[3][1];
		r.z = V.x * m[0][2] + V.y * m[1][2] + V.z * m[2][2] + m[3][2];
		// scale
		float s = m[3][3];
		s += V.x * m[0][3];
		s += V.y * m[1][3];
		s += V.z * m[2][3];
		if(s != 0.0f){
			s = 1/s;
			r.x *= s;
			r.y *= s;
			r.z *= s;
		}
		return r;
	}

	Matrix operator * (Matrix M2) const
	{
		Matrix Ret;
		//Ret.m[0][0] = m[0][0] * M2.m[0][0] + 
		for(int i = 0; i < 4; i++)
			for(int j = 0; j < 4; j++)
			{
				Ret.m[i][j] = 0;
				for(int k = 0; k < 4; k++)
					Ret.m[i][j] += m[k][j] * M2.m[i][k];
			}
		return Ret;
	}

	void SetIdentityMatrix(void)
	{
		for(int i = 0; i < 4; i++)
			for(int j = 0; j < 4; j++)
				m[i][j] = ((i == j) ? 1.0f : 0.0f);
	}

	void SetTranslationMatrix(Vector Translation)
	{
		SetIdentityMatrix();
		m[3][0] = Translation.x;
		m[3][1] = Translation.y;
		m[3][2] = Translation.z;
	}

	void SetScaleTranslationMatrix(Vector Scale, Vector Translation)
	{
		SetIdentityMatrix();
		m[0][0] = Scale.x;
		m[1][1] = Scale.y;
		m[2][2] = Scale.z;
		m[3][0] = Translation.x;
		m[3][1] = Translation.y;
		m[3][2] = Translation.z;
	}

	void SetRotationXMatrix(float Angle)
	{
		SetIdentityMatrix();
		m[1][1] =  (float)cos(Angle);
		m[1][2] =  (float)sin(Angle);
		m[2][2] =  (float)cos(Angle);
		m[2][1] = -(float)sin(Angle);
	}

	void SetRotationYMatrix(float Angle)
	{
		SetIdentityMatrix();
		m[0][0] =  (float)cos(Angle);
		m[0][2] = -(float)sin(Angle);
		m[2][2] =  (float)cos(Angle);
		m[2][0] =  (float)sin(Angle);
	}

	void SetRotationZMatrix(float Angle)
	{
		SetIdentityMatrix();
		m[0][0] =  (float)cos(Angle);
		m[0][1] =  (float)sin(Angle);
		m[1][1] =  (float)cos(Angle);
		m[1][0] = -(float)sin(Angle);
	}

	// Translation:	Position of object
	// Angle:		XZ Angle of object
	// Normal:		Normal of terrain
	void SetTangentMatrix(Vector Translation, float Angle, Vector Normal)
	{
		// TODO: optimize for speed!
		Matrix Mtx;
		Mtx.SetRotationYMatrix(-Angle);
		Normal = Mtx * Normal;

		SetIdentityMatrix();
		Vector z(0, -Normal.z, Normal.y);
		z = Vector::Normalize(z);
		Vector x = Vector::CrossProduct(Normal, z);

		Mtx.SetRotationYMatrix(Angle);
		x = Mtx * x;
		z = Mtx * z;
		Normal = Mtx * Normal;

		m[1][0] = Normal.x;
		m[1][1] = Normal.y;
		m[1][2] = Normal.z;
		m[2][0] = z.x;
		m[2][1] = z.y;
		m[2][2] = z.z;
		m[0][0] = x.x;
		m[0][1] = x.y;
		m[0][2] = x.z;
		m[3][0] = Translation.x;
		m[3][1] = Translation.y;
		m[3][2] = Translation.z;
	
	}
};

}

#endif
