/***********************************************************/
/*                                                         */
/* D3DWindow.cpp                                           */
/*                                                         */
/* created 13.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#include "D3DWindow.h"

// Debug
#include "MessageWindow.h"
#include "LogFile.h"

#include <stdio.h>
#include <stdlib.h>

// www.infinitecode.com 3d library
namespace I3D
{

// lookup table for HWND and D3DWindow
ISL::HashTable<HWND, D3DWindow *, 10, 10> D3DWindow::HWndTable;

D3DWindow::D3DWindow(void)
:HWnd(NULL)
{
	ITL::LogFile::Clear(DBG_D3D);
}

D3DWindow::~D3DWindow(void)
{
	if(HWnd) Destroy();
}

// called by windows to send window messages
LRESULT CALLBACK D3DW_WndProc(HWND HWnd, UINT Message, WPARAM WParam, LPARAM LParam)
{
	D3DWindow *Wnd = D3DWindow::HWndTable.GetValue(HWnd);
	if(Wnd) return Wnd->OnWindowMessage(Message, WParam, LParam);
	return DefWindowProc(HWnd, Message, WParam, LParam);
}

void D3DWindow::Create(int WindowStyle, char *Caption, int dx, int dy)
{
	ITL::LogFile::Msg(DBG_D3D, "D3DWindow::Create...");
	DWCreateWindow(WindowStyle, Caption, dx, dy);
	DWCreateDirect3D();
}

void D3DWindow::DWCreateWindow(int WindowStyle, char *Caption, int dx, int dy)
{
	ITL::LogFile::Msg(DBG_D3D, "D3DWindow::DWCreateWindow...");
	strcpy(WndClassName, "I3D::D3DWindow");
	D3DWindow::Instance = Instance;
	D3DWindow::Pos.WindowStyle = WindowStyle;
	// create Window class
	WinCl.hInstance = Instance;
	WinCl.lpszClassName = WndClassName;
	WinCl.lpfnWndProc = D3DW_WndProc;
	WinCl.style = 0;//CS_NOCLOSE;
	WinCl.hIcon = (HICON)LoadImage(Instance, "Tank4.ico", IMAGE_ICON, 0, 0, LR_LOADFROMFILE);
	WinCl.hCursor = LoadCursor(NULL, IDC_ARROW);
	WinCl.lpszMenuName = NULL;
	WinCl.cbClsExtra = 0;
	WinCl.cbWndExtra = 0;
	WinCl.hbrBackground = NULL;//(HBRUSH) GetStockObject(LTGRAY_BRUSH);
	// set client size
	Pos.Clientdx = dx;
	Pos.Clientdy = dy;
	DesiredPos = Pos;
	// create a window
	if(RegisterClass(&WinCl))
	{
		HWnd = CreateWindowEx(
			/*WS_EX_TOPMOST*/0,WndClassName, Caption, WS_OVERLAPPEDWINDOW, 0, 0, Pos.Clientdx, Pos.Clientdy,
			HWND_DESKTOP, NULL, GetModuleHandle(NULL), NULL);
		ShowWindow(HWnd, WindowStyle);
	}
	else
	{
		throw D3DWException("Can't register window class");
	}
	// get client size
	RECT rect;
	GetClientRect(HWnd, &rect);
	Pos.Clientdx = rect.right - rect.left;
	Pos.Clientdy = rect.bottom - rect.top;
	DesiredPos = Pos;
	// add the window to the hwnd-d3dwindow lookup table
	HWndTable.Add(HWnd, this);
}

void D3DWindow::Destroy(void)
{
	ITL::LogFile::Msg(DBG_D3D, "D3DWindow::Destroy...");
	DWDestroyDirect3D();
	DWDestroyWindow();
}

void D3DWindow::DWDestroyWindow(void)
{
	ITL::LogFile::Msg(DBG_D3D, "D3DWindow::DWDestroyWindow...");
	HWND OldHWnd = HWnd;
	DestroyWindow(HWnd);
	HWnd = NULL;
	HWndTable.DeleteKey(OldHWnd);
	UnregisterClass(WndClassName, Instance);
}

void D3DWindow::ResetDirect3D(void)
{
	ITL::LogFile::Msg(DBG_D3D, "D3DWindow::ResetDirect3D...");
	if(DesiredPos.WindowStyle == SW_SHOWMAXIMIZED)
	{
		ITL::LogFile::Msg(DBG_D3D, "ResetDirect3D: mode: fullscreen");
		// TODO: fill
 	}
	else
	{
		ITL::LogFile::Msg(DBG_D3D, "ResetDirect3D: mode: windowed");
		D3DDISPLAYMODE D3DDM;
		if(D3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &D3DDM) != D3D_OK)
		{
			ITL::LogFile::Msg(DBG_D3D, "ResetDirect3D: Couldn't get displaymode (D3D->GetAdapterDisplayMode)");
			throw D3DWException("ResetDirect3D: Couldn't get displaymode (D3D->GetAdapterDisplayMode)");
		}
		D3DPP.Windowed = true;
		D3DPP.BackBufferFormat = D3DDM.Format;
		D3DPP.BackBufferWidth = DesiredPos.Clientdx;
		D3DPP.BackBufferHeight = DesiredPos.Clientdy;
	}
	D3DPP.AutoDepthStencilFormat = D3DFMT_D16; // TODO: 32 bit ?!?
	static bool b = true;
	if(D3DDevice->Reset(&D3DPP) != D3D_OK)
	{
		if(D3DDevice->TestCooperativeLevel() == D3DERR_DEVICELOST)
			ITL::LogFile::Msg(DBG_D3D, "ResetDirect3D: D3DDevice->Reset returned D3DERR_DEVICELOST");
		ITL::LogFile::Msg(DBG_D3D, "ResetDirect3D: D3DDevice->Reset failed");
		throw D3DWException("ResetDirect3D: D3DDevice->Reset failed");
	}
	Pos = DesiredPos;
	SendEvent(D3DWEvent(D3DW_EVENT_RESET));
}

LRESULT CALLBACK D3DWindow::OnWindowMessage(UINT Message, WPARAM WParam, LPARAM LParam)
{
	switch (Message){
		case WM_CREATE:
			break;
		case WM_DESTROY:
			PostQuitMessage(0);
			break;
		case WM_SIZE:
			switch(WParam)
			{
			case SIZEFULLSCREEN:
				// maximized
				DesiredPos.WindowStyle = SW_SHOWMAXIMIZED;
				// TODO: maximize window
				break;
			case SIZENORMAL:
				DesiredPos.WindowStyle = SW_SHOWNORMAL;
				// TODO: unmaximize window, if maximized
				break;
			case SIZEICONIC:
				break;
			}
			// resize
			DesiredPos.Clientdx = LParam & 0xFFFF;
			DesiredPos.Clientdy = LParam >> 16;
			ResetDirect3D();
			SendEvent(D3DWEvent(D3DW_EVENT_WINDOWSEVENT, Message, WParam, LParam));
			return 0;
			break;
	}
	SendEvent(D3DWEvent(D3DW_EVENT_WINDOWSEVENT, Message, WParam, LParam));
	return DefWindowProc(HWnd, Message, WParam, LParam);
}

void D3DWindow::DWCreateDirect3D(void)
{
	// create a d3d object
	ITL::LogFile::Msg(DBG_D3D, "D3DWindow::DWCreateDirect3D...");
	D3D = Direct3DCreate8(D3D_SDK_VERSION);
	if(!D3D)
	{
		ITL::LogFile::Msg(DBG_D3D, "DWCreateDirect3D: Direct3DCreate8 returned nullpointer");
		throw D3DWException("DWCreateDirect3D: Direct3DCreate8 returned nullpointer");
	}

	// enumerate display modes
	D3DDISPLAYMODE Enum, Mode;
	Mode.Width = 0;
	char mstr[1024];
	for(int i = 0; i < (int)D3D->GetAdapterModeCount(D3DADAPTER_DEFAULT); i++)
	{
		D3D->EnumAdapterModes(D3DADAPTER_DEFAULT, i, &Enum);
		if(Enum.Width == 640 && Enum.Height == 480
		&&Enum.Format == D3DFMT_R5G6B5/* || Enum.Format == D3DFMT_A8R8G8B8 || Enum.Format == D3DFMT_X8R8G8B8)*/)
		{
/*					if(Enum.Width > Mode.Width) Mode = Enum;
					if(Enum.Width < CompMode.Width) CompMode = Enum;
					if(Enum.Width == Mode.Width
						&& (Enum.Format == D3DFMT_A8R8G8B8 || Enum.Format == D3DFMT_X8R8G8B8) && (Mode.Format == D3DFMT_R5G6B5)) Mode = Enum;
					if(Enum.Width == CompMode.Width
						&& (Enum.Format == D3DFMT_R5G6B5) && (CompMode.Format == D3DFMT_A8R8G8B8 || CompMode.Format == D3DFMT_X8R8G8B8)) CompMode = Enum;
					if(Enum.Width == Mode.Width
						&& Enum.Format == Mode.Format
						&& Enum.RefreshRate > Mode.RefreshRate)
						Mode = Enum;
					if(Enum.Width == CompMode.Width
						&& Enum.Format == CompMode.Format
						&& Enum.RefreshRate < CompMode.RefreshRate)
						CompMode = Enum;*/
			Mode = Enum;
			sprintf(mstr, "enum display mode (ok): %dx%d, %d bit, %d Hz", Enum.Width, Enum.Height, (Enum.Format == D3DFMT_R5G6B5)?16:32, Enum.RefreshRate);
			ITL::LogFile::Msg(DBG_D3D, mstr);
		}
		else
		{
			sprintf(mstr, "enum display mode (rejected): %dx%d, %d, %d Hz", Enum.Width, Enum.Height, Enum.Format, Enum.RefreshRate);
			ITL::LogFile::Msg(DBG_D3D, mstr);
		}
	}
	if(Mode.Width == 0)
	{
		ITL::LogFile::Msg(DBG_D3D, "DWCreateDirect3D: Display Mode Enumeration returned no usable display modes");
		throw D3DWException("DWCreateDirect3D: Display Mode Enumeration returned no usable display modes");
	}
	sprintf(mstr, "selected display mode: %dx%d, %d bit, %d Hz", Mode.Width, Mode.Height, (Mode.Format == D3DFMT_R5G6B5)?16:32, Mode.RefreshRate);
	ITL::LogFile::Msg(DBG_D3D, mstr);

	// create a d3d device
	// set D3DPP
	memset(&D3DPP, 0, sizeof(D3DPP));
	D3DPP.BackBufferCount= 1;
	D3DPP.AutoDepthStencilFormat = D3DFMT_D16;
	D3DPP.EnableAutoDepthStencil = true; // z buffer
	D3DPP.SwapEffect = D3DSWAPEFFECT_DISCARD;
	D3DPP.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
	if(Pos.WindowStyle == SW_SHOWMAXIMIZED)
	{
		// fullscreen
		ITL::LogFile::Msg(DBG_D3D, "mode: fullscreen");
		D3DPP.Windowed = false;
		D3DPP.hDeviceWindow  = HWnd;
		D3DPP.BackBufferWidth = Mode.Width;
		D3DPP.BackBufferHeight = Mode.Height;
		D3DPP.BackBufferFormat = Mode.Format;
		D3DPP.FullScreen_RefreshRateInHz = Mode.RefreshRate;
	}
	else
	{
		// windowed
		ITL::LogFile::Msg(DBG_D3D, "mode: windowed");
		D3DDISPLAYMODE D3DDM;
		if(D3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &D3DDM) != D3D_OK)
		{
			ITL::LogFile::Msg(DBG_D3D, "DWCreateDirect3D: Couldn't get displaymode (D3D->GetAdapterDisplayMode)");
			throw D3DWException("DWCreateDirect3D: Couldn't get displaymode (D3D->GetAdapterDisplayMode)");
		}
		D3DPP.Windowed = true;
		D3DPP.BackBufferFormat = D3DDM.Format;
		D3DPP.BackBufferWidth = Pos.Clientdx;
		D3DPP.BackBufferHeight = Pos.Clientdy;
		D3DPP.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
	}

	// create device
	if(D3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, HWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, &D3DPP, &D3DDevice) == D3D_OK)
		ITL::LogFile::Msg(DBG_D3D, "T&L supported. Using hardware vertex processing");
	else if(D3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, HWnd, D3DCREATE_MIXED_VERTEXPROCESSING, &D3DPP, &D3DDevice) == D3D_OK)
		ITL::LogFile::Msg(DBG_D3D, "T&L partially supported. Using mixed vertex processing");
	else if(D3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, HWnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &D3DPP, &D3DDevice) == D3D_OK)
		ITL::LogFile::Msg(DBG_D3D, "T&L not supported. Using software vertex processing");
	else
	{
		ITL::LogFile::Msg(DBG_D3D, "DWCreateDirect3D: Can't create device");
		throw D3DWException("DWCreateDirect3D: Can't create device");
	}
	ITL::LogFile::Msg(DBG_D3D, "CreateDevice successful");
}

void D3DWindow::DWDestroyDirect3D(void)
{
	ITL::LogFile::Msg(DBG_D3D, "D3DWindow::DWDestroyDirect3D...");
	RELEASE(D3DDevice);
	RELEASE(D3D);
	ITL::LogFile::Msg(DBG_D3D, "D3DWindow::DWDestroyDirect3D successful");
}

}
