/***********************************************************/
/*                                                         */
/* Game.cpp                                                */
/*                                                         */
/* created 13.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#include "Game.h"

#include "MathStructures.h"
#include "LogFile.h"
#include "MessageWindow.h"
#include "Text.h"
#include "DataStructures.h"
#include "IMath.h"

// Tank 4
namespace T4
{

void Game::Create(int WindowStyle)
{
	// TODO: implement fullscreen display
	WindowStyle = SW_SHOWNORMAL;
	ShowCursor(false);
	// clear the error output file
	ITL::LogFile::Clear("Error.txt");
	// Create the main window and initialize D3D
	try
	{
		Wnd.Create(WindowStyle, "Tank 4", 640, 480);
	}
	catch(exception& Ex)
	{
		ITL::LogFile::Msg("Error.txt", (char *)Ex.what());
		ITL::MessageWindow::Message((char *)Ex.what(), "Error during D3D Window initialisation");
		exit(-1);
	}
	Wnd.AddListener(this);
	DS.Create(Wnd.D3DDevice, &Sp);
	Init3D();
	LoadingScreen();

	Mouse.Create(&DS, Wnd.GetHWnd());
	Wrld.Create(&DS);
}

void Game::LoadingScreen(void)
{
	DS.D3DDevice->Clear(0, NULL, D3DCLEAR_ZBUFFER | D3DCLEAR_TARGET, 0x00000000, 1.0f, 0);
	D3DMATRIX Matrix = {1,0,0,0,  0,1,0,0,  0,0,1,0,  0,0,0,1};
	// world and view matrix
	DS.D3DDevice->SetTransform(D3DTS_WORLD, &Matrix);
	DS.D3DDevice->SetTransform(D3DTS_VIEW, &Matrix);
	DS.D3DDevice->SetTransform(D3DTS_PROJECTION, &Matrix);
	// loading text
	char Text[1024] = "loading...";
	I3D::Text::DrawString(&DS, Text, -0.05f * 0.5f * strlen(Text), -0.05f, 0.05f, 0.05f);
	// display
	DS.D3DDevice->Present(NULL, NULL, NULL, NULL);
}

void Game::Init3D(void)
{
	DS.D3DDevice->SetRenderState(D3DRS_LIGHTING,TRUE);
	DS.D3DDevice->SetRenderState(D3DRS_ZENABLE,TRUE);
    DS.D3DDevice->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
    DS.D3DDevice->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    DS.D3DDevice->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_POINT);
    DS.D3DDevice->SetTextureStageState(1, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
    DS.D3DDevice->SetTextureStageState(1, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    DS.D3DDevice->SetTextureStageState(1, D3DTSS_MIPFILTER, D3DTEXF_POINT);
	DS.D3DDevice->SetRenderState(D3DRS_CULLMODE,D3DCULL_CW);
	DS.D3DDevice->SetRenderState(D3DRS_AMBIENT, 0x00000000);
//	DS.D3DDevice->SetRenderState(D3DRS_AMBIENT, 0x40404040);
//	DS.D3DDevice->SetRenderState(D3DRS_AMBIENT, 0x80808080);
	// TODO: shut off clipping?
//	DS.D3DDevice->SetRenderState(D3DRS_CLIPPING, false);


	// for debugging:
	//DS.D3DDevice->SetRenderState(D3DRS_AMBIENT, 0xFFFFFFFF);
	if(D3DFlags & GAME_D3DFLAG_WIREFRAME)
		DS.D3DDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);

	D3DMATERIAL8 Material;
	memset(&Material, 0, sizeof(D3DMATERIAL8));
	Material.Diffuse.r = 1.0f;
	Material.Diffuse.g = 1.0f;
	Material.Diffuse.b = 1.0f;
	Material.Diffuse.a = 1.0f;
	Material.Ambient.r = 1.0f;
	Material.Ambient.g = 1.0f;
	Material.Ambient.b = 1.0f;
	Material.Ambient.a = 1.0f;
	Material.Specular.r = 1.0f;// Licht spiegeln
	Material.Specular.g = 1.0f;
	Material.Specular.b = 1.0f;
	Material.Specular.a = 1.0f;
	Material.Power = 0.0f;
	//  Material.power = (float)10.0;
	DS.D3DDevice->SetMaterial(&Material);
	Wrld.Init3D(&DS);
	AsyncReady = true;
}

// called asynchronously!
void Game::OnEvent(I3D::D3DWEvent Event)
{
	if(!AsyncReady) return;
	switch(Event.GetID())
	{
	case I3D::D3DW_EVENT_RESET:
		Init3D();
		break;
	default:
		Events[Events.Count()] = Event;
	}
}

void Game::EventHandler(void)
{
	if(Events.Count())
	{
		for(int i = 0; i < Events.Count(); i++)
		{
			switch(Events[i].GetID())
			{
			/*case I3D::D3DW_EVENT_KEYDOWN:
				KeyboardInput(Events[i].GetInfo());
				break;*/
			case I3D::D3DW_EVENT_WINDOWSEVENT:
				switch(Events[i].GetInfo())
				{
				case WM_KEYDOWN:
					KeyboardHandler(Events[i].GetWParam());
					break;
				/*case WM_LBUTTONDOWN:
					break;
				case WM_MOUSEWHEEL:
					int Wheel = (short)(Events[i].GetWParam() >> 16);
					break;*/
				}
				break;
			}
		}
		Events.Clear();
	}
}

void Game::KeyboardHandler(char c)
{
	switch(c)
	{
	case 27:
		// Escape
		PostQuitMessage(0);
		break;
	case 'w':
	case 'W':
		D3DFlags ^= GAME_D3DFLAG_WIREFRAME;
		if(D3DFlags & GAME_D3DFLAG_WIREFRAME)
			DS.D3DDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);
		else
			DS.D3DDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
		break;
	case 'p':
	case 'P':
		Sp.Paused = !Sp.Paused;
		break;
	case VK_NUMPAD4:
	case VK_LEFT:
		Wrld.MoveCamera(I3D::Vector(-2, 0, 0));
		break;
	case VK_NUMPAD6:
	case VK_RIGHT:
		Wrld.MoveCamera(I3D::Vector(2, 0, 0));
		break;
	case VK_NUMPAD8:
	case VK_UP:
		Wrld.MoveCamera(I3D::Vector(0, 0, 2));
		break;
	case VK_NUMPAD2:
	case VK_DOWN:
		Wrld.MoveCamera(I3D::Vector(0, 0, -2));
		break;
	case VK_NUMPAD7:
	case VK_HOME:
		Wrld.MoveCamera(I3D::Vector(0, 2, 0));
		break;
	case VK_NUMPAD1:
	case VK_END:
		Wrld.MoveCamera(I3D::Vector(0, -2, 0));
		break;
	case VK_ADD:
		Wrld.MoveCameraDirect(I3D::Vector(0, 0, 2));
		break;
	case VK_SUBTRACT:
		Wrld.MoveCameraDirect(I3D::Vector(0, 0, -2));
		break;
	case VK_NUMPAD9:
	case VK_PRIOR:
		Wrld.RotateCamera(I3D::Vector(-0.01f, 0, 0));
		break;
	case VK_NUMPAD3:
	case VK_NEXT:
		Wrld.RotateCamera(I3D::Vector(0.01f, 0, 0));
		break;
	case VK_NUMPAD0:
	case VK_INSERT:
		Wrld.RotateCamera(I3D::Vector(0, -0.01f, 0));
		break;
	case VK_DELETE:
	case VK_DECIMAL:
		Wrld.RotateCamera(I3D::Vector(0, 0.01f, 0));
		break;
	case VK_DIVIDE:
		Wrld.RotateCamera(I3D::Vector(0, 0, 0.01f));
		break;
	case VK_MULTIPLY:
		Wrld.RotateCamera(I3D::Vector(0, 0, -0.01f));
		break;
	}
}

void Game::MouseHandler(void)
{
	Scrolling();
	I3D::MouseCursor::MouseCursorEvent Event;
	if(Mouse.GetEvent(&Event))
	{
		switch(Event.Code)
		{
		case I3D::MC_CLICK:
			Wrld.Click(IMATH::Vector2D(Event.x, Event.y));
			break;
		case I3D::MC_DRAGEND:
			I3D::BoundingBox2D BBox;
			BBox.Data[I3D::BB_LEFT] = minf(Event.x, Event.x2);
			BBox.Data[I3D::BB_RIGHT] = maxf(Event.x, Event.x2);
			BBox.Data[I3D::BB_BOTTOM] = minf(Event.y, Event.y2);
			BBox.Data[I3D::BB_TOP] = maxf(Event.y, Event.y2);
			Wrld.Select(&BBox);
			break;
		}
	}
}

void Game::Scrolling(void)
{
	bool Inside = true;
	if(Mouse.GetX() < -1 || Mouse.GetX() > 1
	|| Mouse.GetY() < -1 || Mouse.GetY() > 1)
		Inside = false;
	if(Inside)
	{
		float d = Sp.Delta;
		if(Mouse.GetX() < -0.9f)
			Wrld.MoveCamera(IMATH::Vector(-48 * d, 0, 0));
		if(Mouse.GetX() > 0.9f)
			Wrld.MoveCamera(IMATH::Vector( 48 * d, 0, 0));
		if(Mouse.GetY() < -0.9f)
			Wrld.MoveCamera(IMATH::Vector(0, 0, -48 * d));
		if(Mouse.GetY() > 0.9f)
			Wrld.MoveCamera(IMATH::Vector(0, 0,  48 * d));
		static float LastZ = Mouse.GetZ();
		float dZ = Mouse.GetZ() - LastZ;
		static float InterpolateddZ = 0;
		InterpolateddZ *= 0.9f;
		InterpolateddZ += 0.1f * dZ;
		if(abs(InterpolateddZ) > 0.001f)
			Wrld.MoveCameraDirect(IMATH::Vector(0, 0, 0.1f * InterpolateddZ));
		LastZ = Mouse.GetZ();
	}
}

void Game::MainLoop(void)
{
	Sp.Update();
	EventHandler();
	MouseHandler();
	Wrld.Physics(&DS);
	if(DS.D3DDevice->BeginScene())
	{
		if(D3DFlags & GAME_D3DFLAG_WIREFRAME)
			DS.D3DDevice->Clear(0, NULL, D3DCLEAR_ZBUFFER | D3DCLEAR_TARGET, 0x00000000, 1.0f, 0);
		else
			DS.D3DDevice->Clear(0, NULL, D3DCLEAR_ZBUFFER | D3DCLEAR_TARGET, 0x0040B0FF, 1.0f, 0);
		D3DMATRIX Matrix = {1,0,0,0,  0,1,0,0,  0,0,1,0,  0,0,0,1};
		// world and view matrix
		DS.D3DDevice->SetTransform(D3DTS_WORLD, &Matrix);
		DS.D3DDevice->SetTransform(D3DTS_VIEW, &Matrix);
		DS.D3DDevice->SetTransform(D3DTS_PROJECTION, &Matrix);
		DS.D3DDevice->GetTransform(D3DTS_VIEW, &DS.ViewMtx);
		// projection matrix
		// distance eye->monitor in meters
		const float EyeDistance = 0.6f;
		// (vertical) monitor size in meters
		const float MonitorSize = 0.33f;
		// perspective projection: [x, y, z] /= (1 + z)
		D3DMATRIX PMatrix = {1,0,0,0,  0,1,0,0,  0,0,1,1,  0,0,0,1};
		DS.D3DDevice->MultiplyTransform(D3DTS_PROJECTION, &PMatrix);
		// adjust to monitor size and position
		// screen 3:4 = 0.75f
		D3DMATRIX TMatrix3 = {0.75f * 2.0f / MonitorSize,0,0,0,  0,2.0f / MonitorSize,0,0,  0,0,1.0f / EyeDistance,0,  0,0,0,1};
		DS.D3DDevice->MultiplyTransform(D3DTS_PROJECTION, &TMatrix3);
		DS.D3DDevice->GetTransform(D3DTS_PROJECTION, &DS.ProjMtx);
		// render
		Wrld.Render(&DS);
		// overlay rendering
		Mouse.Render(&DS);
		// display fps
		if(Sp.DeltaLongTerm)
		{
			char Buf[1024];
			itoa(1.0f / Sp.DeltaLongTerm, Buf, 10);
			strcat(Buf, " fps");
			I3D::Text::DrawString(&DS, Buf, 0.5f, 0.93f, 0.05f, 0.05f);
		}
		// display
		DS.D3DDevice->Present(NULL, NULL, NULL, NULL);
		DS.D3DDevice->EndScene();
	}
}

}
