/***********************************************************/
/*                                                         */
/* QTNode.cpp                                              */
/*                                                         */
/* created 11.08.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#include "QTNode.h"
#include "QuadTree.h"

#include "MathStructures.h"

namespace I3D
{

QTNode::QTNode(TreeCreateStruct *TCS)
{
//	int i;
	Flags = TCS->Flags & QT_CHILD_MASK;
	Parent = TCS->Parent;
	Level = TCS->Level;
	Tree = TCS->Tree;
	if(Level < Tree->Specs.TexDepth)
		Flags |= QT_FLAG_SUBDIV;
	// set bounding box x and z coordinates
	if(TCS->Parent)
	{
		switch(TCS->Flags & QT_CHILD_MASK)
		{
		case QT_CHILD_NW:
			BBox.Data[BB_LEFT] = Parent->BBox.Data[BB_LEFT];
			BBox.Data[BB_RIGHT] = 0.5f*(Parent->BBox.Data[BB_LEFT] + Parent->BBox.Data[BB_RIGHT]);
			BBox.Data[BB_NEAR] = 0.5f*(Parent->BBox.Data[BB_NEAR] + Parent->BBox.Data[BB_FAR]);
			BBox.Data[BB_FAR] = Parent->BBox.Data[BB_FAR];
			break;
		case QT_CHILD_NE:
			BBox.Data[BB_LEFT] = 0.5f*(Parent->BBox.Data[BB_LEFT] + Parent->BBox.Data[BB_RIGHT]);
			BBox.Data[BB_RIGHT] = Parent->BBox.Data[BB_RIGHT];
			BBox.Data[BB_NEAR] = 0.5f*(Parent->BBox.Data[BB_NEAR] + Parent->BBox.Data[BB_FAR]);
			BBox.Data[BB_FAR] = Parent->BBox.Data[BB_FAR];
			break;
		case QT_CHILD_SW:
			BBox.Data[BB_LEFT] = Parent->BBox.Data[BB_LEFT];
			BBox.Data[BB_RIGHT] = 0.5f*(Parent->BBox.Data[BB_LEFT] + Parent->BBox.Data[BB_RIGHT]);
			BBox.Data[BB_NEAR] = Parent->BBox.Data[BB_NEAR];
			BBox.Data[BB_FAR] = 0.5f*(Parent->BBox.Data[BB_NEAR] + Parent->BBox.Data[BB_FAR]);
			break;
		case QT_CHILD_SE:
			BBox.Data[BB_LEFT] = 0.5f*(Parent->BBox.Data[BB_LEFT] + Parent->BBox.Data[BB_RIGHT]);
			BBox.Data[BB_RIGHT] = Parent->BBox.Data[BB_RIGHT];
			BBox.Data[BB_NEAR] = Parent->BBox.Data[BB_NEAR];
			BBox.Data[BB_FAR] = 0.5f*(Parent->BBox.Data[BB_NEAR] + Parent->BBox.Data[BB_FAR]);
			break;
		}
	}
	else
	{
		// root node
		BBox.Data[BB_LEFT] = -0.5f * TCS->Tree->Specs.Size;
		BBox.Data[BB_RIGHT] = 0.5f * TCS->Tree->Specs.Size;
		BBox.Data[BB_NEAR] = -0.5f * TCS->Tree->Specs.Size;
		BBox.Data[BB_FAR] = 0.5f * TCS->Tree->Specs.Size;
	}
	// set bounding box y coordinates
	BBox.Data[BB_BOTTOM] = 0; // these dummy values are replaced later
	BBox.Data[BB_TOP] = 0;
	// outer bounding box for node visibility calculation
	VBox.Data[BB_LEFT] = BBox.Data[BB_LEFT] - 8.0f;
	VBox.Data[BB_RIGHT] = BBox.Data[BB_RIGHT] + 8.0f;
	VBox.Data[BB_NEAR] = BBox.Data[BB_NEAR] - 8.0f;
	VBox.Data[BB_FAR] = BBox.Data[BB_FAR] + 8.0f;
	VBox.Data[BB_BOTTOM] = BBox.Data[BB_BOTTOM] - 8.0f;
	VBox.Data[BB_TOP] = BBox.Data[BB_TOP] + 8.0f;
	// set center
	Center = Vector(0.5f * (BBox.Data[BB_LEFT] + BBox.Data[BB_RIGHT]),
	                0.5f * (BBox.Data[BB_BOTTOM] + BBox.Data[BB_TOP]),
					0.5f * (BBox.Data[BB_NEAR] + BBox.Data[BB_FAR]));
	// create children
	if(Level >= TCS->Tree->Specs.Depth)
	{
		Child[QT_CHILD_NW] = NULL;
		Child[QT_CHILD_NE] = NULL;
		Child[QT_CHILD_SW] = NULL;
		Child[QT_CHILD_SE] = NULL;
	}
	else
	{
		// new level
		TCS->Level++;
		TCS->Parent = this;
		// create new TCSs
		TreeCreateStruct NewTCS[4];
		// copy
		NewTCS[0] = *TCS;
		NewTCS[1] = *TCS;
		NewTCS[2] = *TCS;
		NewTCS[3] = *TCS;
		// modify for subnodes
		NewTCS[0].Flags = QT_CHILD_NW;
		NewTCS[1].Flags = QT_CHILD_NE;
		NewTCS[2].Flags = QT_CHILD_SW;
		NewTCS[3].Flags = QT_CHILD_SE;
		// create subnodes
		if(TCS->Level < TCS->Tree->Specs.TexDepth)
		{
			Child[QT_CHILD_NW] = new QTNode(&NewTCS[0]);
			Child[QT_CHILD_NE] = new QTNode(&NewTCS[1]);
			Child[QT_CHILD_SW] = new QTNode(&NewTCS[2]);
			Child[QT_CHILD_SE] = new QTNode(&NewTCS[3]);
		}
		else if(TCS->Level == TCS->Tree->Specs.TexDepth)
		{
			Child[QT_CHILD_NW] = new QTNodeTerraTex(&NewTCS[0]);
			Child[QT_CHILD_NE] = new QTNodeTerraTex(&NewTCS[1]);
			Child[QT_CHILD_SW] = new QTNodeTerraTex(&NewTCS[2]);
			Child[QT_CHILD_SE] = new QTNodeTerraTex(&NewTCS[3]);
		}
		else if(TCS->Level > TCS->Tree->Specs.TexDepth)
		{
			Child[QT_CHILD_NW] = new QTNodeTerra(&NewTCS[0]);
			Child[QT_CHILD_NE] = new QTNodeTerra(&NewTCS[1]);
			Child[QT_CHILD_SW] = new QTNodeTerra(&NewTCS[2]);
			Child[QT_CHILD_SE] = new QTNodeTerra(&NewTCS[3]);
		}
	}
}

void QTNode::ClearSubdivisions(void)
{
	if(Flags & QT_FLAG_SUBDIV)
	{
		Flags &= ~QT_FLAG_SUBDIV;
		for(int i = 0; i < 4; i++)
			if(Child[i])
				Child[i]->ClearSubdivisions();
	}
}

void QTNode::Visibility(D3DStruct *DS)
{
	if(Level < Tree->Specs.TexDepth)
	{
		CalculateScreenBBox(DS);
		Flags |= QT_FLAG_SUBDIV;
	}
	for(int i = 0; i < 4; i++)
		if(Child[i])
			Child[i]->Visibility(DS);
}

void QTNode::BalanceSubTrees(D3DStruct *DS)
{
	for(int i = 0; i < 4; i++)
		if(Child[i])
			Child[i]->BalanceSubTrees(DS);
}

void QTNode::CreateScene(D3DStruct *DS)
{
	for(int i = 0; i < 4; i++)
		if(Child[i])
			Child[i]->CreateScene(DS);
}

void QTNode::CreateTerrain(void)
{
	using namespace IMATH;
	int i;
	// neighbors
	// north
	Vector2DXZ Pos = Vector2DXZ(0.5f * (BBox.Data[BB_LEFT] + BBox.Data[BB_RIGHT])
	                          , 1.5f * BBox.Data[BB_FAR] - 0.5f * BBox.Data[BB_NEAR]);
	Neighbors[QT_NEIGHBOR_N] = Tree->GetNode(&Pos, Level);
	// south
	Pos = Vector2DXZ(0.5f * (BBox.Data[BB_LEFT] + BBox.Data[BB_RIGHT])
	               , 1.5f * BBox.Data[BB_NEAR] - 0.5f * BBox.Data[BB_FAR]);
	Neighbors[QT_NEIGHBOR_S] = Tree->GetNode(&Pos, Level);
	// west
	Pos = Vector2DXZ(1.5f * BBox.Data[BB_LEFT] - 0.5f * BBox.Data[BB_RIGHT]
	               , 0.5f * (BBox.Data[BB_NEAR] + BBox.Data[BB_FAR]));
	Neighbors[QT_NEIGHBOR_W] = Tree->GetNode(&Pos, Level);
	// east
	Pos = Vector2DXZ(1.5f * BBox.Data[BB_RIGHT] - 0.5f * BBox.Data[BB_LEFT]
	               , 0.5f * (BBox.Data[BB_NEAR] + BBox.Data[BB_FAR]));
	Neighbors[QT_NEIGHBOR_E] = Tree->GetNode(&Pos, Level);
	// children
	for(i = 0; i < 4; i++)
		if(Child[i])
			Child[i]->CreateTerrain();
	// BBox, VBox
	if(Level < Tree->Specs.Depth)
	{
		BBox.Data[BB_BOTTOM] = Child[0]->BBox.Data[BB_BOTTOM];
		BBox.Data[BB_TOP] = Child[0]->BBox.Data[BB_TOP];
		for(i = 0; i < 4; i++)
		{
			if(Child[i]->BBox.Data[BB_BOTTOM] < BBox.Data[BB_BOTTOM])
				BBox.Data[BB_BOTTOM] = Child[i]->BBox.Data[BB_BOTTOM];
			if(Child[i]->BBox.Data[BB_TOP] > BBox.Data[BB_TOP])
				BBox.Data[BB_TOP] = Child[i]->BBox.Data[BB_TOP];
		}
		VBox.Data[BB_LEFT] = BBox.Data[BB_LEFT] - 8.0f;
		VBox.Data[BB_RIGHT] = BBox.Data[BB_RIGHT] + 8.0f;
		VBox.Data[BB_NEAR] = BBox.Data[BB_NEAR] - 8.0f;
		VBox.Data[BB_FAR] = BBox.Data[BB_FAR] + 8.0f;
		VBox.Data[BB_BOTTOM] = BBox.Data[BB_BOTTOM] - 8.0f;
		VBox.Data[BB_TOP] = BBox.Data[BB_TOP] + 8.0f;
		// set center
		Center = Vector(0.5f * (BBox.Data[BB_LEFT] + BBox.Data[BB_RIGHT]),
						0.5f * (BBox.Data[BB_BOTTOM] + BBox.Data[BB_TOP]),
						0.5f * (BBox.Data[BB_NEAR] + BBox.Data[BB_FAR]));
	}
}

QTNode *QTNode::GetNode(IMATH::Vector2DXZ *Pos, int Level)
{
	if(BBox.IsIn(Pos))
	{
		if(Level == QTNode::Level)
			return this;
		else
			for(int i = 0; i < 4; i++)
				if(Child[i]->BBox.IsIn(Pos))
					return Child[i]->GetNode(Pos, Level);
	}
	return NULL;
}

void QTNode::AddRenderableModel(RenderableModel *Obj)
{
	for(int i = 0; i < 4; i++)
	{
		IMATH::Vector2DXZ Pos = IMATH::Vector2DXZ(Obj->Pos.x, Obj->Pos.z);
		if(Child[i]->BBox.IsIn(&Pos))
			Child[i]->AddRenderableModel(Obj);
	}
}

void QTNode::DeleteRenderableModel(RenderableModel *Obj)
{
	for(int i = 0; i < 4; i++)
	{
		IMATH::Vector2DXZ Pos = IMATH::Vector2DXZ(Obj->Pos.x, Obj->Pos.z);
		if(Child[i]->BBox.IsIn(&Pos))
			Child[i]->DeleteRenderableModel(Obj);
	}
}

void QTNode::CalculateScreenBBox(D3DStruct *DS)
{
	// outline vectors of the bounding box
	Vector Vec[8];
	Vec[0] = Vector(BBox.Data[BB_LEFT ], BBox.Data[BB_BOTTOM], BBox.Data[BB_NEAR]);
	Vec[1] = Vector(BBox.Data[BB_RIGHT], BBox.Data[BB_BOTTOM], BBox.Data[BB_NEAR]);
	Vec[2] = Vector(BBox.Data[BB_LEFT ], BBox.Data[BB_TOP   ], BBox.Data[BB_NEAR]);
	Vec[3] = Vector(BBox.Data[BB_RIGHT], BBox.Data[BB_TOP   ], BBox.Data[BB_NEAR]);
	Vec[4] = Vector(BBox.Data[BB_LEFT ], BBox.Data[BB_BOTTOM], BBox.Data[BB_FAR ]);
	Vec[5] = Vector(BBox.Data[BB_RIGHT], BBox.Data[BB_BOTTOM], BBox.Data[BB_FAR ]);
	Vec[6] = Vector(BBox.Data[BB_LEFT ], BBox.Data[BB_TOP   ], BBox.Data[BB_FAR ]);
	Vec[7] = Vector(BBox.Data[BB_RIGHT], BBox.Data[BB_TOP   ], BBox.Data[BB_FAR ]);
	Vector Low, High;
	for(int i = 0; i < 8; i++)
	{
		//Vec[i] = DS->ViewMtx * Vec[i];
		Vec[i] = DS->ProjViewMtx * Vec[i];
		if(Vec[i].z < 0 || Vec[i].z > 1)
		{
			//Vec[i].z = 1.0f;
			Vec[i].x = -Vec[i].x;
			Vec[i].y = -Vec[i].y;
		}
		if(i == 0)
		{
			// initialize
			Low = Vector(Vec[i].x, Vec[i].y, Vec[i].z);
			High = Vector(Vec[i].x, Vec[i].y, Vec[i].z);
		}
		if(Vec[i].x < Low.x) Low.x = Vec[i].x;
		if(Vec[i].x > High.x) High.x = Vec[i].x;
		if(Vec[i].y < Low.y) Low.y = Vec[i].y;
		if(Vec[i].y > High.y) High.y = Vec[i].y;
		if(Vec[i].z < Low.z) Low.z = Vec[i].z;
		if(Vec[i].z > High.z) High.z = Vec[i].z;
	}
	ScreenBBox = BoundingBox(Low.x, High.x, Low.y, High.y, Low.z, High.z);
}

XZPos QTNode::GetXZPos(IMATH::Vector2D Pos)
{
	XZPos Ret;
	if(ScreenBBox.IsIn(&Pos) || (ScreenBBox.Data[BB_NEAR] < 0 && ScreenBBox.Data[BB_FAR] > 0))
	{
		XZPos ChRet[4];
		int HighestLevel = 0, HighestNum = -1;
		for(int i = 0; i < 4; i++)
			if(Child[i])
			{
				ChRet[i] = Child[i]->GetXZPos(Pos);
				if(ChRet[i].Level > HighestLevel)
				{
					HighestLevel = ChRet[i].Level;
					HighestNum = i;
				}
			}
		if(HighestNum != -1 && HighestLevel > 0)
		{
			return ChRet[HighestNum];
		}
		else
		{
			Ret.Level = Level;
			Ret.Pos = IMATH::Vector2DXZ(Center.x, Center.z);
			return Ret;
		}
	}
	else
	{
		Ret.Pos = IMATH::Vector2DXZ(0, 0);
		Ret.Level = 0;
		return Ret;
	}
}

QTNode::~QTNode(void)
{
	for(int i = 0; i < 4; i++)
		if(Child[i])
			delete Child[i];
}

// ****** QTNodeRoot ******

QTNodeRoot::QTNodeRoot(TreeCreateStruct *TCS)
:QTNode(TCS)
{
	// we are still in the constructor, so Tree->Root has not been set yet
	Tree->Root = this;
	QTNode::CreateTerrain(); // can't call virtual functions of this object in constructor!
}

void QTNodeRoot::Visibility(D3DStruct *DS)
{
	ClearSubdivisions();
	QTNode::Visibility(DS);
	BalanceSubTrees(DS);
	CreateScene(DS);
}

// ****** QTNodeTerra ******

QTNodeTerra::QTNodeTerra(TreeCreateStruct *TCS)
:QTNode(TCS)
{
}

// calculate terrain complexity for detail-based LODing
void QTNodeTerra::CalculateComplexity(void)
{
	// calculate node complexity for detail-based LODing
	Vector NS = 0.5f * (TexNode->Vertices[Indices[QT_INDEX_N]].v + TexNode->Vertices[Indices[QT_INDEX_S]].v);
	Vector WE = 0.5f * (TexNode->Vertices[Indices[QT_INDEX_W]].v + TexNode->Vertices[Indices[QT_INDEX_E]].v);
	float Diff1 = (NS - TexNode->Vertices[Indices[QT_INDEX_C]].v).Length();
	float Diff2 = (WE - TexNode->Vertices[Indices[QT_INDEX_C]].v).Length();
	Complexity = max(Diff1, Diff2);
	Complexity /= NS.Length();
//	Complexity = Complexity * Complexity;
	// subnode complexity
	if(Level != Tree->Specs.Depth)
		Complexity = max(Complexity, max(((QTNodeTerra *)Child[QT_CHILD_SW])->Complexity,
		                             max(((QTNodeTerra *)Child[QT_CHILD_SE])->Complexity,
									 max(((QTNodeTerra *)Child[QT_CHILD_NW])->Complexity,
									     ((QTNodeTerra *)Child[QT_CHILD_NE])->Complexity))));
}

void QTNodeTerra::CalculateVisibility(D3DStruct *DS)
{
	Flags &= ~QT_FLAG_VISIBLE_MASK;
	// parent is completely visible -> this node is completely visible
	if(Level != Tree->Specs.TexDepth && (Parent->Flags & QT_FLAG_VISIBLE_COMPLETELY))
	{
		Flags |= (QT_FLAG_VISIBLE_COMPLETELY | QT_FLAG_VISIBLE_PARTLY);
		return;
	}
	// parent not visible -> this node is not visible
	if(Level != Tree->Specs.TexDepth && !(Parent->Flags & QT_FLAG_VISIBLE_PARTLY))
		return;
	// outline vectors of the visibility box
	Vector Vec[8];
	Vec[0] = Vector(VBox.Data[BB_LEFT ], VBox.Data[BB_BOTTOM], VBox.Data[BB_NEAR]);
	Vec[1] = Vector(VBox.Data[BB_RIGHT], VBox.Data[BB_BOTTOM], VBox.Data[BB_NEAR]);
	Vec[2] = Vector(VBox.Data[BB_LEFT ], VBox.Data[BB_TOP   ], VBox.Data[BB_NEAR]);
	Vec[3] = Vector(VBox.Data[BB_RIGHT], VBox.Data[BB_TOP   ], VBox.Data[BB_NEAR]);
	Vec[4] = Vector(VBox.Data[BB_LEFT ], VBox.Data[BB_BOTTOM], VBox.Data[BB_FAR ]);
	Vec[5] = Vector(VBox.Data[BB_RIGHT], VBox.Data[BB_BOTTOM], VBox.Data[BB_FAR ]);
	Vec[6] = Vector(VBox.Data[BB_LEFT ], VBox.Data[BB_TOP   ], VBox.Data[BB_FAR ]);
	Vec[7] = Vector(VBox.Data[BB_RIGHT], VBox.Data[BB_TOP   ], VBox.Data[BB_FAR ]);
	// all edges of the visibility box
	const int Edges[12][2] = {	{0, 1}, {0, 2}, {0, 4},
								{5, 1}, {5, 4}, {5, 7},
								{3, 1}, {3, 2}, {3, 7},
								{6, 2}, {6, 7}, {6, 4}};
	//
	// Code[8]:
	//
	// code representing the position of the vertex on the screen:
	//
	//           1001 |  1000  | 1010
	//          ------+--------+------
	//           0001 |  0000  | 0010
	//                | screen |
	//          ------+--------+------
	//           0101 |  0100  | 0110
	//
	// vertices with the code 0000 are inside the screen clipping area [-1.0f, 1.0f]
	// additionally (because of 3D), a fifth digit 10000 is added, if Vertex.z is < -1.0f
	//
	// if (Code[x] & Code[y]) is 0, the box is potentially partially visible
	// if (Code[x] | Code[y]) is 0, the box is completely visible
	//
	int Code[8];
	int i;
	// calculate the code
	for(i = 0; i < 8; i++)
	{
		Code[i] = 0;
		// TODO: find bug with a visible line at ca. z = 75
		Vec[i] = DS->ProjViewMtx * Vec[i];
		//if(Vec[i].z <= 0.90f) Code[i] |= 16;
		//if(Vec[i].z <= 0.99f) Code[i] |= 16;
		if(Vec[i].x < -1.0f) Code[i] |= 1;
		if(Vec[i].x >  1.0f) Code[i] |= 2;
		if(Vec[i].y < -1.0f) Code[i] |= 4;
		if(Vec[i].y >  1.0f) Code[i] |= 8;
		if(Vec[i].z >  1.0f)
		{
			Code[i] |= 16; // this means z < -1 before the projection matrix is applied
/*			Vec[i].x = -Vec[i].x;
			Vec[i].y = -Vec[i].y;*/
		}
	}
	Flags |= QT_FLAG_VISIBLE_COMPLETELY;
	for(i = 0; i < 12; i++)
	{
		// test if the box is partly or completely visible
		if(!(Code[Edges[i][0]] & Code[Edges[i][1]]))
			Flags |= QT_FLAG_VISIBLE_PARTLY;
		if(Code[Edges[i][0]] | Code[Edges[i][1]])
			Flags &= ~QT_FLAG_VISIBLE_COMPLETELY;
	}
}

#define QT_COMPLEXITY_MIN		0.01f
#define QT_COMPLEXITY_MIN_SQRT	0.1f
#define QT_COMPLEXITY_DEPTH		1200.0f
#define QT_COMPLEXITY_DEPTH_SQ	(QT_COMPLEXITY_DEPTH * QT_COMPLEXITY_DEPTH)

void QTNodeTerra::Visibility(D3DStruct *DS)
{
	Flags &= ~QT_FLAG_SUBDIV;

	CalculateVisibility(DS);
	CalculateScreenBBox(DS);
	if(Flags & QT_FLAG_VISIBLE_PARTLY)
	{
		// calculate distance from camera
		Vector Screen = Center;
		Screen = DS->ViewMtx * Screen;
		Distance = Screen.Length();
		// calculate distance factor for terrain LODing
		if(Distance >= 0 && Distance <  100.0f) Distance =  100.0f;
		if(Distance <  0 && Distance > -100.0f) Distance = -100.0f;
		if(Distance != 0)
			Distance = QT_COMPLEXITY_DEPTH / Distance;
		// subdivide
		if(Complexity * Distance >= QT_COMPLEXITY_MIN && Level != Tree->Specs.Depth)
		{
			Flags |= QT_FLAG_SUBDIV;
			QTNode::Visibility(DS);
		}
	}
}

void QTNodeTerra::BalanceSubTrees(D3DStruct *DS)
{
	if(Flags & QT_FLAG_SUBDIV)
	{
		ShowNeighbors(DS);
		QTNode::BalanceSubTrees(DS);
	}
}

void QTNodeTerra::ShowNeighbors(D3DStruct *DS)
{
	if(Level > Tree->Specs.TexDepth)
		for(int i = 0; i < 4; i++)
			if(Neighbors[i])
				((QTNodeTerra *)Neighbors[i])->Show(DS);
}

void QTNodeTerra::Show(D3DStruct *DS)
{
	if(Level > Tree->Specs.TexDepth)
		if(!(Parent->Flags & QT_FLAG_SUBDIV))
			if(Parent->Flags & QT_FLAG_VISIBLE_PARTLY)
			{
				// subdivide parent
				Parent->Flags |= QT_FLAG_SUBDIV;
				// visibility
				for(int i = 0; i < 4; i++)
					((QTNodeTerra *)Parent->Child[i])->CalculateVisibility(DS);
				// show parent and parent's neighbors
				((QTNodeTerra *)Parent)->ShowNeighbors(DS);
				((QTNodeTerra *)Parent)->Show(DS);
			}
}

void QTNodeTerra::AddRenderableModel(RenderableModel *Obj)
{
	// TODO: fix bug with actor visibility
	Objects[Objects.Count()] = Obj;
	if((Flags & QT_FLAG_VISIBLE_PARTLY) && !(Flags & QT_FLAG_SUBDIV))
		((QuadTree::ObjectInformation *)Obj->GetData())->Visible = true;
	if(Level == Tree->Specs.Depth)
		((QuadTree::ObjectInformation *)Obj->GetData())->DeepestNode = this;
	for(int i = 0; i < 4; i++)
	{
		IMATH::Vector2DXZ Pos = IMATH::Vector2DXZ(Obj->Pos.x, Obj->Pos.z);
		if(Child[i] && Child[i]->BBox.IsIn(&Pos))
			Child[i]->AddRenderableModel(Obj);
	}
}

void QTNodeTerra::DeleteRenderableModel(RenderableModel *Obj)
{
	if(Objects.Delete(Obj))
	{
		for(int i = 0; i < 4; i++)
		{
			if(Child[i])
				Child[i]->DeleteRenderableModel(Obj);
		}
	}
}

void QTNodeTerra::CreateScene(D3DStruct *DS)
{
/*	if(!(Flags & QT_FLAG_VISIBLE) && (Flags & QT_FLAG_SUBDIV))
		Flags += 0;*/
	if(Flags & QT_FLAG_VISIBLE_PARTLY)
	{
		if(!(Flags & QT_FLAG_SUBDIV))
		{
			// objects
			for(int i = 0; i < Objects.Count(); i++)
				Tree->AddVisibleObject(Objects[i]);
			// terrain
			bool North = Neighbors[QT_NEIGHBOR_N] && (Neighbors[QT_NEIGHBOR_N]->Flags & QT_FLAG_SUBDIV);
			bool South = Neighbors[QT_NEIGHBOR_S] && (Neighbors[QT_NEIGHBOR_S]->Flags & QT_FLAG_SUBDIV);
			bool West  = Neighbors[QT_NEIGHBOR_W] && (Neighbors[QT_NEIGHBOR_W]->Flags & QT_FLAG_SUBDIV);
			bool East  = Neighbors[QT_NEIGHBOR_E] && (Neighbors[QT_NEIGHBOR_E]->Flags & QT_FLAG_SUBDIV);
			if(!(North || South || East || West) && Complexity * Distance < QT_COMPLEXITY_MIN)
			{
				// low detail quad
				TexNode->RS.AddIndex(Indices[QT_INDEX_NW]);
				TexNode->RS.AddIndex(Indices[QT_INDEX_SW]);
				TexNode->RS.AddIndex(Indices[QT_INDEX_NE]);
				TexNode->RS.AddIndex(Indices[QT_INDEX_NE]);
				TexNode->RS.AddIndex(Indices[QT_INDEX_SW]);
				TexNode->RS.AddIndex(Indices[QT_INDEX_SE]);
			}
			else
			{
				// north
				if(North)
				{
					// two triangles, so no gaps occur
					TexNode->RS.AddIndex(Indices[QT_INDEX_NW]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_C]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_N]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_N]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_C]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_NE]);
				}
				else
				{
					// one triangle
					TexNode->RS.AddIndex(Indices[QT_INDEX_NW]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_C]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_NE]);
				}
				// east
				if(East)
				{
					// two triangles, so no gaps occur
					TexNode->RS.AddIndex(Indices[QT_INDEX_NE]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_C]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_E]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_E]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_C]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_SE]);
				}
				else
				{
					// one triangle
					TexNode->RS.AddIndex(Indices[QT_INDEX_NE]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_C]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_SE]);
				}
				// south
				if(South)
				{
					// two triangles, so no gaps occur
					TexNode->RS.AddIndex(Indices[QT_INDEX_SE]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_C]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_S]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_S]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_C]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_SW]);
				}
				else
				{
					// one triangle
					TexNode->RS.AddIndex(Indices[QT_INDEX_SE]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_C]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_SW]);
				}
				// west
				if(West)
				{
					// two triangles, so no gaps occur
					TexNode->RS.AddIndex(Indices[QT_INDEX_SW]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_C]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_W]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_W]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_C]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_NW]);
				}
				else
				{
					// one triangle
					TexNode->RS.AddIndex(Indices[QT_INDEX_SW]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_C]);
					TexNode->RS.AddIndex(Indices[QT_INDEX_NW]);
				}
			}
		}
		else
		{
			QTNode::CreateScene(DS);
		}
	}
}

void QTNodeTerra::CreateTerrain(void)
{
	// get QTNodeTerraTex
	QTNodeTerra *Srch = this;
	while(Srch->Level > Tree->Specs.TexDepth)
		Srch = (QTNodeTerra *)Srch->Parent;
	TexNode = (QTNodeTerraTex *)Srch;
	if(Level > Tree->Specs.TexDepth)
	{
		QTNodeTerra *Parent = (QTNodeTerra *)QTNodeTerra::Parent;
		switch(Flags & QT_CHILD_MASK)
		{
		case QT_CHILD_NW:
			Indices[QT_INDEX_SW] = Parent->Indices[QT_INDEX_W];
			Indices[QT_INDEX_SE] = Parent->Indices[QT_INDEX_C];
			Indices[QT_INDEX_NW] = Parent->Indices[QT_INDEX_NW];
			Indices[QT_INDEX_NE] = Parent->Indices[QT_INDEX_N];
			break;
		case QT_CHILD_NE:
			Indices[QT_INDEX_SW] = Parent->Indices[QT_INDEX_C];
			Indices[QT_INDEX_SE] = Parent->Indices[QT_INDEX_E];
			Indices[QT_INDEX_NW] = Parent->Indices[QT_INDEX_N];
			Indices[QT_INDEX_NE] = Parent->Indices[QT_INDEX_NE];
			break;
		case QT_CHILD_SW:
			Indices[QT_INDEX_SW] = Parent->Indices[QT_INDEX_SW];
			Indices[QT_INDEX_SE] = Parent->Indices[QT_INDEX_S];
			Indices[QT_INDEX_NW] = Parent->Indices[QT_INDEX_W];
			Indices[QT_INDEX_NE] = Parent->Indices[QT_INDEX_C];
			break;
		case QT_CHILD_SE:
			Indices[QT_INDEX_SW] = Parent->Indices[QT_INDEX_S];
			Indices[QT_INDEX_SE] = Parent->Indices[QT_INDEX_SE];
			Indices[QT_INDEX_NW] = Parent->Indices[QT_INDEX_C];
			Indices[QT_INDEX_NE] = Parent->Indices[QT_INDEX_E];
			break;
		}
		Indices[QT_INDEX_S] = (Indices[QT_INDEX_SW] + Indices[QT_INDEX_SE]) / 2;
		Indices[QT_INDEX_W] = (Indices[QT_INDEX_NW] + Indices[QT_INDEX_SW]) / 2;
		Indices[QT_INDEX_E] = (Indices[QT_INDEX_NE] + Indices[QT_INDEX_SE]) / 2;
		Indices[QT_INDEX_N] = (Indices[QT_INDEX_NW] + Indices[QT_INDEX_NE]) / 2;
		Indices[QT_INDEX_C] = (Indices[QT_INDEX_N ] + Indices[QT_INDEX_S ]) / 2;
	}

	QTNode::CreateTerrain();

	// BBox, VBox
	if(Level == Tree->Specs.Depth)
	{
		BBox.Data[BB_BOTTOM] = TexNode->Vertices[Indices[QT_INDEX_SW]].y;
		BBox.Data[BB_TOP] = TexNode->Vertices[Indices[QT_INDEX_SW]].y;
		for(int i = 0; i < 9; i++)
		{
			if(TexNode->Vertices[Indices[i]].y < BBox.Data[BB_BOTTOM])
				BBox.Data[BB_BOTTOM] = TexNode->Vertices[Indices[i]].y;
			if(TexNode->Vertices[Indices[i]].y > BBox.Data[BB_TOP])
				BBox.Data[BB_TOP] = TexNode->Vertices[Indices[i]].y;
		}
		VBox.Data[BB_LEFT] = BBox.Data[BB_LEFT] - 8.0f;
		VBox.Data[BB_RIGHT] = BBox.Data[BB_RIGHT] + 8.0f;
		VBox.Data[BB_NEAR] = BBox.Data[BB_NEAR] - 8.0f;
		VBox.Data[BB_FAR] = BBox.Data[BB_FAR] + 8.0f;
		VBox.Data[BB_BOTTOM] = BBox.Data[BB_BOTTOM] - 8.0f;
		VBox.Data[BB_TOP] = BBox.Data[BB_TOP] + 8.0f;
		// set center
		Center = Vector(0.5f * (BBox.Data[BB_LEFT] + BBox.Data[BB_RIGHT]),
						0.5f * (BBox.Data[BB_BOTTOM] + BBox.Data[BB_TOP]),
						0.5f * (BBox.Data[BB_NEAR] + BBox.Data[BB_FAR]));
	}
	// Complexity
	CalculateComplexity();
}

// ****** QTNodeTerraTex ******

QTNodeTerraTex::QTNodeTerraTex(TreeCreateStruct *TCS)
:QTNodeTerra(TCS)
{
}

void QTNodeTerraTex::CreateVertices(QuadTreeSpecs *QTSpecs)
{
//	for(int j = VtxSize - 1; j >= 0; j--) // z
	for(int j = 0; j < VtxSize; j++) // z
		for(int i = 0; i < VtxSize; i++) // x
		{
			float XFac = (float)i / (VtxSize - 1);
			float ZFac = (float)j / (VtxSize - 1);
			float x = (1 - XFac) * BBox.Data[BB_LEFT] + XFac * BBox.Data[BB_RIGHT];
			float z = (1 - ZFac) * BBox.Data[BB_NEAR] + ZFac * BBox.Data[BB_FAR];
//			float z = (1 - ZFac) * BBox.Data[BB_FAR] + ZFac * BBox.Data[BB_NEAR];
			Vertices[j * VtxSize + i] = VertexTex2(Vector(x, QTSpecs->Surface->GetY(x, z), z), QTSpecs->Surface->GetNormal(x, z), Vector2D(XFac, 1 - ZFac), Vector2D(16 * XFac, 16 * (1 - ZFac)));
			RS.AddVertex(Vertices[j * VtxSize + i]);
		}
	RS.CreateVBuf(QTSpecs->DS->D3DDevice);
}

void QTNodeTerraTex::CreateView(D3DStruct *DS)
{
}

void QTNodeTerraTex::Render(D3DStruct *DS)
{
	RenderableIndexTex2::Render(DS);
}

void QTNodeTerraTex::CreateScene(D3DStruct *DS)
{
	Tree->AddRenderable(this);
	RS.ClearIndex();

	QTNodeTerra::CreateScene(DS);

	RS.CreateIBuf(Tree->Specs.DS->D3DDevice);
}

void QTNodeTerraTex::CreateTerrain(void)
{
	// create vertices
	VtxSize = (1 << (Tree->Specs.Depth - Tree->Specs.TexDepth + 1)) + 1;
	Vertices = new VertexTex2[VtxSize * VtxSize];
	CreateVertices(&Tree->Specs);
	// create indices
	int Max = VtxSize - 1;
	int Mid = Max / 2;
	Indices[QT_INDEX_SW] = 0;
	Indices[QT_INDEX_S ] = Mid;
	Indices[QT_INDEX_SE] = Max;
	Indices[QT_INDEX_W ] = Mid * VtxSize;
	Indices[QT_INDEX_C ] = Mid * VtxSize + Mid;
	Indices[QT_INDEX_E ] = Mid * VtxSize + Max;
	Indices[QT_INDEX_NW] = Max * VtxSize;
	Indices[QT_INDEX_N ] = Max * VtxSize + Mid;
	Indices[QT_INDEX_NE] = Max * VtxSize + Max;

	QTNodeTerra::CreateTerrain();
}

QTNodeTerraTex::~QTNodeTerraTex()
{
	delete[] Vertices;
}

}
