/***********************************************************/
/*                                                         */
/* QTNode.h                                                */
/*                                                         */
/* created 11.08.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef QTNODE_H
#define QTNODE_H

#include "D3DStructures.h"
#include "DataStructures.h"
#include "Renderable.h"
#include "RenderableModel.h"
#include "DynamicArray.h"

// www.infinitecode.com 3d library
namespace I3D
{

// flag and index values
enum
{
	// index indices
	QT_INDEX_SW = 0,
	QT_INDEX_S  = 1,
	QT_INDEX_SE = 2,
	QT_INDEX_W  = 3,
	QT_INDEX_C  = 4,
	QT_INDEX_E  = 5,
	QT_INDEX_NW = 6,
	QT_INDEX_N  = 7,
	QT_INDEX_NE = 8,
	// neighbor direction indices
	QT_NEIGHBOR_N = 0,
	QT_NEIGHBOR_S = 1,
	QT_NEIGHBOR_W = 2,
	QT_NEIGHBOR_E = 3,
	// child indices (used in TreeCreateStruct::Flags and QTNode::Flags)
	QT_CHILD_SW = 0,
	QT_CHILD_SE = 1,
	QT_CHILD_NW = 2,
	QT_CHILD_NE = 3,
	QT_CHILD_MASK = 3,
	// QTNode flags
	QT_FLAG_SUBDIV =				4,
	QT_FLAG_VISIBLE_PARTLY =		8,
	QT_FLAG_VISIBLE_COMPLETELY =	16,
	QT_FLAG_VISIBLE_MASK =			16 + 8
};

// in QuadTree.h:
class QuadTree;
struct QuadTreeSpecs;
class XZPos;

class QTNode
{
public:
	struct TreeCreateStruct
	{
		int Level;
		QTNode *Parent;
		int Flags;
		QuadTree *Tree;
	};

	int Flags;
	int Level;
	QTNode *Parent;
	QTNode *Child[4];
	QTNode *Neighbors[4];
	QuadTree *Tree;
	BoundingBox BBox, VBox, ScreenBBox; // Bounding box, outer bbox, and screen bbox for node visibility calculation
	Vector Center;

	QTNode(TreeCreateStruct *TCS);
	virtual ~QTNode(void);

	// create terrain vertices and indices
	virtual void CreateTerrain(void);
	// clear subdivision flags
	void ClearSubdivisions(void);
	// calculate terrain visibility for CreateScene
	virtual void Visibility(D3DStruct *DS);
	// balance subtrees for CreateScene
	virtual void BalanceSubTrees(D3DStruct *DS);
	// add visible indices and objects to the quadtree
	virtual void CreateScene(D3DStruct *DS);
	// add a model to the node
	virtual void AddRenderableModel(RenderableModel *Obj);
	virtual void DeleteRenderableModel(RenderableModel *Obj);

	// get the node at a given position in a given level
	QTNode *GetNode(IMATH::Vector2DXZ *Pos, int Level);
	void CalculateScreenBBox(D3DStruct *DS);
	XZPos GetXZPos(IMATH::Vector2D Pos);
};

// ****** QTNodeRoot ******

class QTNodeRoot: public QTNode
{
public:
	QTNodeRoot(TreeCreateStruct *TCS);
	//virtual ~QTNodeRoot() {}

	// calculate terrain visibility
	virtual void Visibility(D3DStruct *DS);
};

// ****** QTNodeTerra ******

class QTNodeTerraTex;

class QTNodeTerra: public QTNode
{
	// terrain
	QTNodeTerraTex *TexNode;
	float Complexity, Distance;
	// subobjects
	ISL::DynamicArray<RenderableModel *, 0, 1> Objects;

	void ShowNeighbors(D3DStruct *DS);
	void Show(D3DStruct *DS);

protected:
	unsigned short Indices[9];

public:
	QTNodeTerra(TreeCreateStruct *TCS);

	// create terrain vertices and indices
	virtual void CreateTerrain(void);
	// calculate terrain visibility for CreateScene
	virtual void Visibility(D3DStruct *DS);
	// balance subtrees for CreateScene
	virtual void BalanceSubTrees(D3DStruct *DS);
	// add visible indices and objects to the quadtree
	virtual void CreateScene(D3DStruct *DS);
	// add a model to the node
	virtual void AddRenderableModel(RenderableModel *Obj);
	virtual void DeleteRenderableModel(RenderableModel *Obj);
	// calculate terrain complexity for detailbased LODing
	void CalculateComplexity(void);
	// calculate visibility for culling
	void CalculateVisibility(D3DStruct *DS);
};

// ****** QTNodeTerraTex ******

class QTNodeTerraTex: public QTNodeTerra, RenderableIndexTex2
{
	friend QTNodeTerra;

	LPDIRECT3DTEXTURE8 Texture;
	// terrain vertices
	VertexTex2 *Vertices;
	int VtxSize;

	void CreateVertices(QuadTreeSpecs *QTSpecs);

protected:
	virtual void CreateView(D3DStruct *DS);

public:
	virtual void Render(D3DStruct *DS);

public:
	QTNodeTerraTex(TreeCreateStruct *TCS);
	~QTNodeTerraTex(void);

	// create terrain vertices and indices
	virtual void CreateTerrain(void);
	// add visible indices and objects to the quadtree
	virtual void CreateScene(D3DStruct *DS);
};

}

#endif
