/***********************************************************/
/*                                                         */
/* QuadTree.cpp                                            */
/*                                                         */
/* created 16.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

// roam quadtree

#include "QuadTree.h"
#include "QTNode.h"
#include "TextureManager.h"

// www.infinitecode.com 3d library
namespace I3D
{

QuadTree::QuadTree(void)
:Root(NULL)
{
}

QuadTree::~QuadTree(void)
{
	if(Root)
		delete Root;
}

void QuadTree::Create(QuadTreeSpecs *Specs)
{
	QuadTree::Specs = *Specs;
	QTNode::TreeCreateStruct TCS;
	TCS.Level = 0;
	TCS.Parent = NULL;
	TCS.Flags = 0;
	TCS.Tree = this;
	Root = new QTNodeRoot(&TCS);
	//Root->CreateTerrain();
	// create vertices
	//RS.ClearVertex();
	//CreateVertices(Root);
	//RS.CreateVBuf(Specs->DS->D3DDevice);
	// load texture
	//Texture = TextureManager::Load(Specs->DS->D3DDevice, "Textures\\Terrain.bmp", NULL, 0);
	Texture = TextureManager::GetTexture(Specs->DS, "Terrain");
	DetailTexture = TextureManager::GetTexture(Specs->DS, "Textures\\TerrainDetail.bmp", 0, TM_FLAG_GRAYMIPMAPS);
	// create view
	OnViewChange();
}

void QuadTree::CreateView(D3DStruct *DS)
{
	ObjectsRnd.Clear();
	VisibleObjects.Clear();
	Root->Visibility(DS);
	for(int i = 0; i < ObjectsRnd.Count(); i++)
		ObjectsRnd[i]->OnViewChange();
}

void QuadTree::Render(D3DStruct *DS)
{
	Renderable::Render(DS); // calls CreateView when necessary
	// render terrain
	DS->D3DDevice->SetTexture(0, Texture);
	DS->D3DDevice->SetTexture(1, DetailTexture);
	int i;
	DS->D3DDevice->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_ADDSIGNED);
	for(i = 0; i < ObjectsRnd.Count(); i++)
		ObjectsRnd[i]->Render(DS);
	DS->D3DDevice->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
	// render managed objects
	// TODO: optimize dereferences
	// for textures
	for(i = 0; i < VisibleObjects.CountKeys(); i++)
	{
		DS->D3DDevice->SetTexture(0, VisibleObjects.GetKey(i));
		// for models
		for(int j = 0; j < VisibleObjects.GetValuePtr(i)->CountKeys(); j++)
		{
			VisibleObjects.GetValuePtr(i)->GetKey(j)->PreRender(DS);
			VisibleObjects.GetValuePtr(i)->GetKey(j)->RenderSetup(DS);
			for(int k = 0; k < VisibleObjects.GetValuePtr(i)->GetValuesPtr(j)->Count(); k++)
			{
				//(*VisibleObjects.GetValuePtr(i)->GetValuesPtr(j))[k]->Render(DS);
				DS->D3DDevice->SetTransform(D3DTS_WORLD, (*VisibleObjects.GetValuePtr(i)->GetValuesPtr(j))[k]->Mtx);
				DS->D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, VisibleObjects.GetValuePtr(i)->GetKey(j)->RS.GetVertexNumber(), 0, VisibleObjects.GetValuePtr(i)->GetKey(j)->RS.GetIndexNumber() / 3);
			}
			VisibleObjects.GetValuePtr(i)->GetKey(j)->PostRender(DS);
		}
	}
}

void QuadTree::AddRenderable(Renderable *Obj)
{
	ObjectsRnd[ObjectsRnd.Count()] = Obj;
}

void QuadTree::AddRenderableModel(RenderableModel *Obj)
{
	Obj->SetData(new ObjectInformation());
	Root->AddRenderableModel(Obj);
}

void QuadTree::DeleteRenderableModel(RenderableModel *Obj)
{
	if(Obj->GetData()) delete Obj->GetData();
	Obj->SetData(0);
	Root->DeleteRenderableModel(Obj);
}

void QuadTree::MoveRenderableModel(RenderableModel *Obj)
{
	// TODO: fix bug with actor visibility
	IMATH::Vector2DXZ Point = IMATH::Vector2DXZ(Obj->Pos.x, Obj->Pos.z);
	ObjectInformation *Info = (ObjectInformation *)Obj->GetData();
	if((!Info->DeepestNode) || (!Info->DeepestNode->BBox.IsIn(&Point)))
	{
		bool OldVisible = ((ObjectInformation *)Obj->GetData())->Visible;
		DeleteRenderableModel(Obj);
		AddRenderableModel(Obj);
		//QTNode *NewNode = ((ObjectInformation *)Obj->GetData())->DeepestNode;
		if(!((ObjectInformation *)Obj->GetData())->Visible && OldVisible)
		{
			//DeleteVisibleObject(Obj);
		}
		if(((ObjectInformation *)Obj->GetData())->Visible && !OldVisible)
		{
			//AddVisibleObject(Obj);
		}
	}
}

void QuadTree::AddVisibleObject(RenderableModel *Obj)
{
	if(VisibleObjects.GetValuePtr(Obj->Texture))
		VisibleObjects.GetValuePtr(Obj->Texture)->Add(Obj->Mdl, Obj);
	else
		VisibleObjects.Add(Obj->Texture, NodeModel(Obj));
}

void QuadTree::DeleteVisibleObject(RenderableModel *Obj)
{
	NodeModel *NM = VisibleObjects.GetValuePtr(Obj->Texture);
	if(NM)
	{
		NodeModel::Entry::HTEValues *HV = NM->GetValuesPtr(Obj->Mdl);
		if(HV)
			HV->Delete(Obj);
	}
}

QTNode *QuadTree::GetNode(IMATH::Vector2DXZ *Pos, int Level)
{
/*	if(Pos->x < Root->BBox.Data[BB_LEFT ]) Pos->x = Root->BBox.Data[BB_LEFT ];
	if(Pos->x > Root->BBox.Data[BB_RIGHT]) Pos->x = Root->BBox.Data[BB_RIGHT];
	if(Pos->z < Root->BBox.Data[BB_FAR  ]) Pos->z = Root->BBox.Data[BB_FAR  ];
	if(Pos->z > Root->BBox.Data[BB_NEAR ]) Pos->z = Root->BBox.Data[BB_NEAR ];*/
	return Root->GetNode(Pos, Level);
}

XZPos QuadTree::GetXZPos(IMATH::Vector2D Pos)
{
	return Root->GetXZPos(Pos);
}

}
