/***********************************************************/
/*                                                         */
/* QuadTree.h                                              */
/*                                                         */
/* created 16.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

// roam quadtree based on a NURBS surface

#ifndef QUADTREE_H
#define QUADTREE_H

#include "DynamicArray.h"
#include "HashTable.h"
#include "MultiHashTable.h"
#include "D3DStructures.h"
#include "RenderStruct.h"
#include "DataStructures.h"
#include "NURBS.h"
#include "Renderable.h"
#include "RenderableModel.h"
#include "Model.h"

// www.infinitecode.com 3d library
namespace I3D
{

// in QTNode.h:
class QTNode;
class QTNodeRoot;
class QTNodeTerra;
class QTNodeTerraTex;

// model node
/*class QTRNodeModel
{
	Model *Mdl;
	ISL::DynamicArray<RenderableModel *, 0, 1> Obj;
};

// texture node
class QTRNodeTexture
{
	LPDIRECT3DTEXTURE8 Texture;
	ISL::DynamicArray<QTRNodeModel, 0, 2> Models;
};

// RenderableModel Objects
class QTRObjectArray
{
	ISL::DynamicArray<QTRNodeTexture, 0, 8> Objects;
};*/

struct QuadTreeSpecs
{
	int TexDepth, Depth;
	float Size; // size of the HeightMap and Render Size
	NURBS *Surface;
	D3DStruct *DS;
};

class XZPos
{
public:
	IMATH::Vector2DXZ Pos;
	int Level;
};

class QuadTree: public Renderable
{
	friend QTNode;
	friend QTNodeRoot;
	friend QTNodeTerra;
	friend QTNodeTerraTex;

	int Flags;
	// root of the tree structure
	QTNode *Root;

	// TODO: replace by mission textures
	LPDIRECT3DTEXTURE8 Texture, DetailTexture;

	// renderable objects
	ISL::DynamicArray<Renderable *, 0, 8> ObjectsRnd;

	// renderable objects with a model (array of visible objects)
	class NodeModel: public ISL::MultiHashTable<I3D::Model *, I3D::RenderableModel *>
	{
	public:
		NodeModel(void) {};

		NodeModel(I3D::RenderableModel *Obj)
		{
			Add(Obj->Mdl, Obj);
		}
	};
	class NodeTexture: public ISL::HashTable<LPDIRECT3DTEXTURE8, NodeModel> {};
	NodeTexture VisibleObjects;

	//QTRObjectArray Objects;

protected:
	virtual void CreateView(D3DStruct *DS);

public:
	class ObjectInformation
	{
	public:
		ObjectInformation(void)
			:DeepestNode(0), Visible(false)
			{}

		QTNode *DeepestNode;
		bool Visible;
	};

	// the properties of this quadtree
	QuadTreeSpecs Specs;

	QuadTree(void);
	~QuadTree(void);

	virtual void Render(D3DStruct *DS);
	void Create(QuadTreeSpecs *Specs);
	// add an object to tha tree
	void AddRenderable(Renderable *Obj);
	// add an object with a model to the tree
	void AddRenderableModel(RenderableModel *Obj);
	void DeleteRenderableModel(RenderableModel *Obj);
	void MoveRenderableModel(RenderableModel *Obj);
	// add an object to the object array (array of visible objects)
	void AddVisibleObject(RenderableModel *Obj);
	void DeleteVisibleObject(RenderableModel *Obj);
	QTNode *GetNode(IMATH::Vector2DXZ *Pos, int Level);
	XZPos GetXZPos(IMATH::Vector2D Pos);
};

}

#endif
