/***********************************************************/
/*                                                         */
/* TextureLoader.h                                         */
/*                                                         */
/* created 24.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef TEXTUREMANAGER_H
#define TEXTUREMANAGER_H

#include "D3DException.h"
#include "D3DStructures.h"
#include "IniLoader.h"
#include "DynamicArray.h"
#include "D3DStruct.h"

#include <d3d8.h>
#include <stdio.h>

// www.infinitecode.com 3d library
namespace I3D
{

enum
{
	// create alphamap with colorkey 0x000000
	TM_FLAG_ALPHACC0 = 1,
	// optimize alpha color keyed textures for bilinear filtering
	TM_FLAG_ACCBILOPT = 2,
	// ignore texels with alpha = 0 when calculation mipmaps
	TM_FLAG_MIPALPHAIGN = 4,
	// don't allow nonzero alpha < 0x60 in high-level mipmaps
	TM_FLAG_MIPALPHA60 = 8,
	// load flags from Textures\\Textures,ini
	// TODO: implement
	TM_FLAG_FLAGSFROMINI = 16,
	// make high mipmaps gray (used for detail texturing)
	TM_FLAG_GRAYMIPMAPS = 32
};

class TextureManager
{
	class LoadedTexture
	{
	public:
		LoadedTexture(void)
			:Name(0)
		{
		}

		LoadedTexture(const char *Name, LPDIRECT3DTEXTURE8 Texture)
		{
			LoadedTexture::Name = new char[1024];
			strcpy(LoadedTexture::Name, Name);
			LoadedTexture::Texture = Texture;
		}

		~LoadedTexture(void)
		{
			if(Name) delete[] Name;
		}

		LoadedTexture& operator = (LoadedTexture &LT)
		{
			if(!Name)
				Name = new char[1024];
			strcpy(Name, LT.Name);
			Texture = LT.Texture;
			return *this;
		}

		LoadedTexture(LoadedTexture& LT)
		{
			*this = LT;
		}

		char *Name;
		LPDIRECT3DTEXTURE8 Texture;

	};

	// static class, so no instances are allowed
	TextureManager(void) {}

	// creates all mipmap levels for the texture
	static void TextureManager::CreateMipMaps(LPDIRECT3DTEXTURE8 Texture, int Flags)
	{
		int i;
		// get Texture width and height
		D3DSURFACE_DESC LDesc;
		if(Texture->GetLevelDesc(0, &LDesc) != D3D_OK)
			throw TextureManagerException("CreateMipMaps: Can't get description of Texture level 0");
		// create temporary Textureture map
		unsigned int *Map = (unsigned int *)malloc(LDesc.Width * LDesc.Height * sizeof(unsigned int));
		// get the Texture data
		D3DLOCKED_RECT LRect;
		if(Texture->LockRect( 0, &LRect, NULL, 0 )!=D3D_OK)
			throw TextureManagerException("CreateMipMaps: Can't lock Texture");
		unsigned int *TexMap = (unsigned int*)LRect.pBits;
		if(LRect.pBits == NULL)
			throw TextureManagerException("CreateMipMaps: Texture lock returned nullpointer");
		// copy Texture data into temporary map
		for(i = 0; i < (int)LDesc.Width; i++)
		{
			for(int j = 0; j < (int)LDesc.Height; j++)
			{
				Map[((LDesc.Height - 1) - j) * LDesc.Width + i] = TexMap[((LDesc.Height - 1) - j) * LDesc.Width + i];
			}
		}
		if(Texture->UnlockRect(0) != D3D_OK)
			throw TextureManagerException("CreateMipMaps: Can't unlock Texture");
		// create mipmap levels
		int Res = LDesc.Width;
		for(i = 1; i < (int)Texture->GetLevelCount(); i++)
		{
			Res /= 2;
			if(Texture->LockRect(i, &LRect, NULL, 0) != D3D_OK)
				throw TextureManagerException("CreateMipMaps: Can't lock texture mipmap level");
			unsigned int *TexMap = (unsigned int*)LRect.pBits;
			if(LRect.pBits == NULL)
				throw TextureManagerException("CreateMipMaps: Textureture mipmap level lock returned nullpointer");
			int j, k;
			// for all pixels
			for(j = 0; j < Res; j++)
			{
				for(k = 0; k < Res; k++)
				{
					int num = 0;
					int a = 0, r = 0, g = 0, b = 0;
					// for the four pixels of the lower mipmap
					for(int l = 0; l < 2; l++)
						for(int m = 0; m < 2; m++)
						{
							int MapAlpha = Map[(2 * j + m) * LDesc.Width + 2 * k + l] >> 24;
							if(!(Flags & TM_FLAG_MIPALPHAIGN) || MapAlpha)
							{
								// add pixel color
								num++;
								a += MapAlpha;
								r += (Map[(2 * j + m) * LDesc.Width + 2 * k + l] >> 16) & 0xFF;
								g += (Map[(2 * j + m) * LDesc.Width + 2 * k + l] >> 8) & 0xFF;
								b += Map[(2 * j + m) * LDesc.Width + 2 * k + l] & 0xFF;
							}
						}
					if(num)
					{
						// calculate average
						a /= num;
						r /= num;
						g /= num;
						b /= num;
						if(Flags & TM_FLAG_MIPALPHA60 && a && a < 0x60)
							a = 0x60;
						if(Flags & TM_FLAG_GRAYMIPMAPS && i > 3)
						{
							r = (int)((r - 128.0f) * 0.5f + 128.0f);
							g = (int)((g - 128.0f) * 0.5f + 128.0f);
							b = (int)((b - 128.0f) * 0.5f + 128.0f);
						}
					}
					Map[j * LDesc.Width + k] = a * 0x1000000 + r * 0x10000 + g * 0x100 + b;
				}
			}
			// copy the mipmap
			for(j = 0; j < Res; j++)
			{
				for(k = 0; k < Res; k++)
				{
					TexMap[j * Res + k] = Map[j * LDesc.Width + k];
				}
			}
			if(Texture->UnlockRect(i)!=D3D_OK)
				throw TextureManagerException("CreateMipMaps: Can't unlock texture mipmap level");
		}
		delete Map;
	}

	// fills black colorkeyed pixels with neighboring colors
	static void TextureManager::BilinearOptimization(unsigned int *Map, int Width, int Height)
	{
		bool Change;
		do
		{
			Change = false;
			// for all pixels
			for(int i = 0; i < Width; i++)
				for(int j = 0; j < Height; j++)
					// if the pixel is colorkeyed (alpha == 0) and black
					if(Map[j * Width + i] >> 24 == 0 && Map[j * Width + i] & 0xFFFFFF)
					{
						Change = true;
						// initialise color and number of averaged pixels
						int r = 0, b = 0, g = 0;
						int num = 0;
						// for all neighbors
						for(int k = -1; k <= 1; k++)
							for(int l = -1; l <= 1; l++)
								// clipping
								if(i + k >= 0 && i + k < Width && j + l >= 0 && j + l < Height)
								{
									// if not colorkeyed
									if(Map[(j + l) * Width + i + k] >> 24 != 0)
									{
										// store the color
										num++;
										r += (Map[(j + l) * Width + i + k] >> 16) & 0xFF;
										g += (Map[(j + l) * Width + i + k] >> 8) & 0xFF;
										b += Map[(j + l) * Width + i + k] & 0xFF;
									}
								}
						// if a color has been found
						if(num)
						{
							// calculate average
							r /= num;
							g /= num;
							b /= num;
							// paint the pixel
							Map[j * Width + i] = 0x10000 * r + 0x100 * g + b;
						}
					}
		} while(Change);
	}

public:
	class TextureManagerException: public D3DException
	{
	public:
		TextureManagerException(char *Description)
			:D3DException(Description)
		{
			AddPrefix("Texture Manager: ");
		}
	};

	static ITL::IniLoader Ini;

	static ISL::DynamicArray<LoadedTexture, 0, 1> LoadedTextures;

	// loads a texture from a bmp file (Direct3D X functions are evil ;-))
	static LPDIRECT3DTEXTURE8 TextureManager::Load(LPDIRECT3DDEVICE8 D3DDevice, const char *FileName, const char *AlphaFileName, int Flags)
	{
/*		char FName[1024];
		strcpy(FName, "Textures\\");
		strcat(FName, FileName);*/
		char Str[1024];
		LPDIRECT3DTEXTURE8 Texture = NULL;
		FILE *f = fopen(FileName, "rb");
		fseek(f, 18, SEEK_SET);
		int Width, Height, Bit;
		fread(&Width, 4, 1, f);
		fread(&Height, 4, 1, f);
		fseek(f, 28, SEEK_SET);
		fread(&Bit, 4, 1, f);
		D3DDevice->CreateTexture(Width, Height, 0, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &Texture);
		D3DLOCKED_RECT LRect;
		if(Texture->LockRect( 0, &LRect, NULL, 0 )!=D3D_OK)
		{
			sprintf(Str, "TextureLoader::Load(\"%s\") can't lock texture", FileName);
			throw TextureManagerException(Str);
		}
		unsigned int *Buf=(unsigned int*)LRect.pBits;
		unsigned int *Map = (unsigned int *)malloc(Width * Height * sizeof(unsigned int));
		
		if(!Buf)
			throw TextureManagerException("TextureManager::Load: texture lock returned nullpointer");
		switch(Bit)
		{
		case 24:
			{
				fseek(f, -Width*Height*3, SEEK_END);
				// bmp's are BGR and reverse-y, textures are RGB, so we sort everything here
				for(int j = Height - 1; j >= 0; j--)
					for(int i = 0; i < Width; i++)
					{
						int a = 0xFF;
						int b = fgetc(f);
						int g = fgetc(f);
						int r = fgetc(f);
						if(Flags & TM_FLAG_ALPHACC0 && r + b + g == 0)
							a = 0;
						Map[j * Width + i] = 0x1000000 * a + 0x10000 * r + 0x100 * g + b;
					}
				break;
			}
		case 8:
			// TODO: implement this
			break;
		default:
			char Str[1024];
			sprintf(Str, "TextureLoader::Load(\"%s\") invalid bitrate: %d", FileName, Bit);
			RELEASE(Texture);
			throw TextureManagerException(Str);
			exit(-1);
			break;
		}
		if(Flags & TM_FLAG_ACCBILOPT) BilinearOptimization(Map, Width, Height);
		memcpy(Buf, Map, Width * Height * 4);
		delete Map;
		if(Texture->UnlockRect(0) != D3D_OK)
		{
			sprintf(Str, "TextureLoader::Load(\"%s\") can't unlock texture", FileName);
			throw TextureManagerException(Str);
		}
		fclose(f);
		CreateMipMaps(Texture, Flags);
		return Texture;
	}

	// loads a texture with the parameters in Textures\\Textures.ini
	static LPDIRECT3DTEXTURE8 GetTexture(D3DStruct *DS, const char *Name, const char *AlphaName = 0, int Flags = 0)
	{
		// if already loaded, return texture handle
		for(int i = 0; i < LoadedTextures.Count(); i++)
			if(!strcmpi(Name, LoadedTextures[i].Name))
				return LoadedTextures[i].Texture;
		char FileName[1024], AlphaFileName[1024];
		// filename
		const char *IniFileName = Ini.GetValue(Name, "File");
		if(IniFileName)
			strcpy(FileName, IniFileName);
		else
			strcpy(FileName, Name);
		// alpha map filename
		const char *IniAlphaFileName = Ini.GetValue(Name, "AlphaFile");
		if(IniAlphaFileName)
		{
			strcpy(AlphaFileName, "Textures\\");
			strcat(FileName, IniAlphaFileName);
		}
		else
			*AlphaFileName = 0;
		// flags
		// alpha color key 0
		const char *AlphaColorKeyZero = Ini.GetValue(Name, "AlphaColorKeyZero");
		if(AlphaColorKeyZero && !strcmpi(AlphaColorKeyZero, "true"))
			Flags |= TM_FLAG_ALPHACC0;
		// bilinear optimizations for alpha color key
		const char *AlphaCCBilinearOptimize = Ini.GetValue(Name, "AlphaCCBilinearOptimize");
		if(AlphaCCBilinearOptimize && !strcmpi(AlphaCCBilinearOptimize, "true"))
			Flags |= TM_FLAG_ACCBILOPT;
		// alpha = 0 ignore for mipmap creation
		const char *MipMapIgnoreAlpha0 = Ini.GetValue(Name, "MipMapIgnoreAlpha0");
		if(MipMapIgnoreAlpha0 && !strcmpi(MipMapIgnoreAlpha0, "true"))
			Flags |= TM_FLAG_MIPALPHAIGN;
		// keep nonzero alpha >= 60
		const char *MipMapAlpha60 = Ini.GetValue(Name, "MipMapAlpha60");
		if(MipMapAlpha60 && !strcmpi(MipMapAlpha60, "true"))
			Flags |= TM_FLAG_MIPALPHA60;
		const char *GrayMipMaps = Ini.GetValue(Name, "GrayMipMaps");
		if(GrayMipMaps && !strcmpi(GrayMipMaps, "true"))
			Flags |= TM_FLAG_GRAYMIPMAPS;
		LPDIRECT3DTEXTURE8 Texture;
		if(*AlphaFileName)
			Texture = Load(DS->D3DDevice, FileName, AlphaFileName, Flags);
		else
			Texture = Load(DS->D3DDevice, FileName, NULL, Flags);
		LoadedTextures[LoadedTextures.Count()] = LoadedTexture(Name, Texture);
		return Texture;
	}

};

}

#endif
