/***********************************************************/
/*                                                         */
/* World.cpp                                               */
/*                                                         */
/* created 16.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#include "World.h"

#include "IMath.h"
#include "DataStructures.h"
#include "QuadTree.h"
#include "D3DStructures.h"
#include "Tank.h"

#include <math.h>

// www.infinitecode.com game library
namespace IGAME
{

void World::Create(I3D::D3DStruct *DS)
{
	CameraPos = I3D::Vector(0, 200.0f, 0);
	CameraAngle = I3D::Vector(0.20f * PI, 0, 0);
	Init3D(DS);

	const float Size = 1024.0f; // 1 kilometer ;-)

	Surface.Create(256, 256, Size);
	//For.Create(DS, I3D::BoundingBox2D(-0.5f * Size, 0.5f * Size, -0.5f * Size, 0.5f * Size), &Surface);

	// create QuadTree
	I3D::QuadTreeSpecs Specs;
	Specs.TexDepth = 3;
	Specs.Depth = 7;
	Specs.Size = Size;
	Specs.Surface = &Surface;
	Specs.DS = DS;
	QTree.Create(&Specs);

	int i;
	for(i = 0; i < GAME_TREES; i++)
	{
		float x = (rand() % 16384 - 8192) / 16384.0f * Size;
		float z = (rand() % 16384 - 8192) / 16384.0f * Size;
		IGAME::ActorSpecs Specs;
		Specs.DS = DS;
		Specs.Pos = IMATH::Vector(x, Surface.GetY(x, z), z);
		Specs.QT = &QTree;
		StaticActors[i] = new Tree();
		StaticActors[i]->Create(&Specs);
	}
	for(i = 0; i < 256; i++)
	{
		float x = (rand() % 16384 - 8192) / 16384.0f * Size;
		float z = (rand() % 16384 - 8192) / 16384.0f * Size;
		IGAME::ActorSpecs Specs;
		Specs.Pos = IMATH::Vector(x, Surface.GetY(x, z), z);
		Specs.DS = DS;
		Specs.QT = &QTree;
		Actors[i] = new Tank();
		Actors[i]->Create(&Specs);
	}
}

void World::Destroy(void)
{
	int i;
	for(i = 0; i < StaticActors.Count(); i++)
	{
		delete StaticActors[i];
	}
	for(i = 0; i < Actors.Count(); i++)
	{
		delete Actors[i];
	}
}

// called when the camera moves
void World::OnViewChange(void)
{
	RenderFlags |= RENDERABLE_FLAG_VIEWCHANGE;
}

void World::Physics(I3D::D3DStruct *DS)
{
	for(int i = 0; i < Actors.Count(); i++)
	{
		Actors[i]->Physics(DS);
	}
}

void World::CreateView(I3D::D3DStruct *DS)
{
	QTree.OnViewChange();
	//For.OnViewChange();
}

void World::Render(I3D::D3DStruct *DS)
{
	if(RenderFlags & RENDERABLE_FLAG_VIEWCHANGE)
	{
		RenderFlags &= ~RENDERABLE_FLAG_VIEWCHANGE;
		// create new View
		CreateView(DS);
	}
	// set camera
//	Angle += 0.02f;
	D3DMATRIX MatrixZ = {(float)cos(-CameraAngle.z),(float)sin(-CameraAngle.z),0,0,  -(float)sin(-CameraAngle.z),(float)cos(-CameraAngle.z),0,0,  0,0,1,0,  0,0,0,1};
    DS->D3DDevice->MultiplyTransform(D3DTS_VIEW, &MatrixZ);
	D3DMATRIX MatrixX = {1,0,0,0,  0,(float)cos(-CameraAngle.x),(float)sin(-CameraAngle.x),0,  0,-(float)sin(-CameraAngle.x),(float)cos(-CameraAngle.x),0,  0,0,0,1};
    DS->D3DDevice->MultiplyTransform(D3DTS_VIEW, &MatrixX);
	D3DMATRIX MatrixY = {(float)cos(-CameraAngle.y),0,-(float)sin(-CameraAngle.y),0,  0,1,0,0,  (float)sin(-CameraAngle.y),0,(float)cos(-CameraAngle.y),0,  0,0,0,1};
    DS->D3DDevice->MultiplyTransform(D3DTS_VIEW, &MatrixY);
	D3DMATRIX MatrixT = {1,0,0,0,  0,1,0,0,  0,0,1,0,  -CameraPos.x,-CameraPos.y,-CameraPos.z,1};
    DS->D3DDevice->MultiplyTransform(D3DTS_VIEW, &MatrixT);
	DS->D3DDevice->GetTransform(D3DTS_VIEW, &DS->ViewMtx);
	DS->CreateProjViewMtx();
	// debug: screen clipping
	/*D3DMATRIX TMatrix2 = {1,0,0,0,  0,1,0,0,  0,0,1,0,  0,0,800.0f,1};
    DS->D3DDevice->SetTransform(D3DTS_VIEW, &TMatrix2);
    DS->D3DDevice->MultiplyTransform(D3DTS_VIEW, &DS->ViewMtx);*/
	// render
	// quadtree
	QTree.Render(DS);
	// overlay: current selection
	for(int i = 0; i < Selection.Count(); i++)
	{
		Selection[i]->RenderSelection(DS);
	}
}

void World::Init3D(I3D::D3DStruct *DS)
{
	D3DLIGHT8 Light;
	memset(&Light, 0, sizeof(D3DLIGHT8));
	Light.Type = D3DLIGHT_DIRECTIONAL;
	Light.Diffuse.r = 1.0f;
	Light.Diffuse.g = 0.25f;
	Light.Diffuse.b = 0.0f;
	Light.Specular.r=Light.Diffuse.r;
	Light.Specular.g=Light.Diffuse.g;
	Light.Specular.b=Light.Diffuse.b;
	Light.Attenuation0 = 0.0;
	Light.Attenuation1 = 0.0;
	Light.Attenuation2 = 0.0;
	Light.Range = 0;
	Light.Position.x = 0;
	Light.Position.y = 0;
	Light.Position.z = 0;
	Light.Direction.x = -1.0f;
	Light.Direction.y = -1.5f;
	Light.Direction.z = -1.0f;
	DS->D3DDevice->SetLight(1, &Light);
	DS->D3DDevice->LightEnable(1, true);

	memset(&Light, 0, sizeof(D3DLIGHT8));
	Light.Type = D3DLIGHT_DIRECTIONAL;
	Light.Diffuse.r = 0.0f;
	Light.Diffuse.g = 0.25f;
	Light.Diffuse.b = 1.0f;
	Light.Specular.r=Light.Diffuse.r;
	Light.Specular.g=Light.Diffuse.g;
	Light.Specular.b=Light.Diffuse.b;
	Light.Attenuation0 = 0.0;
	Light.Attenuation1 = 0.0;
	Light.Attenuation2 = 0.0;
	Light.Range = 0;
	Light.Position.x = 0;
	Light.Position.y = 0;
	Light.Position.z = 0;
	Light.Direction.x = 1.0f;
	Light.Direction.y = -1.5f;
	Light.Direction.z = 1.0f;
	DS->D3DDevice->SetLight(2, &Light);
	DS->D3DDevice->LightEnable(2, true);

	memset(&Light, 0, sizeof(D3DLIGHT8));
	Light.Type = D3DLIGHT_DIRECTIONAL;
	Light.Diffuse.r = 0.25f;
	Light.Diffuse.g = 0.75f;
	Light.Diffuse.b = 0.25f;
	Light.Specular.r=Light.Diffuse.r;
	Light.Specular.g=Light.Diffuse.g;
	Light.Specular.b=Light.Diffuse.b;
	Light.Attenuation0 = 0.0;
	Light.Attenuation1 = 0.0;
	Light.Attenuation2 = 0.0;
	Light.Range = 0;
	Light.Position.x = 0;
	Light.Position.y = 0;
	Light.Position.z = 0;
	Light.Direction.x = 0.0f;
	Light.Direction.y = -1.0f;
	Light.Direction.z = 0.0f;
	DS->D3DDevice->SetLight(3, &Light);
	DS->D3DDevice->LightEnable(3, true);
}

void World::MoveCamera(I3D::Vector V)
{
	I3D::Matrix Rot;
	Rot.SetRotationYMatrix(CameraAngle.y);
	V = Rot * V;
	CameraPos += V;
	OnViewChange();
}

void World::MoveCameraDirect(I3D::Vector V)
{
	I3D::Matrix Rot;
	Rot.SetRotationZMatrix(CameraAngle.z);
	V = Rot * V;
	Rot.SetRotationXMatrix(CameraAngle.x);
	V = Rot * V;
	Rot.SetRotationYMatrix(CameraAngle.y);
	V = Rot * V;
	CameraPos += V;
	OnViewChange();
}

void World::RotateCamera(I3D::Vector V)
{
	CameraAngle += V;
	OnViewChange();
}

void World::Select(I3D::BoundingBox2D *BBox)
{
	Selection.Clear();
	for(int i = 0; i < Actors.Count(); i++)
	{
		if(Actors[i]->Select(BBox))
		{
			// add to select list
			Selection[Selection.Count()] = Actors[i];
		}
	}
}

void World::Click(IMATH::Vector2D Pos)
{
	I3D::XZPos Target = QTree.GetXZPos(Pos);
	if(Target.Level)
		for(int i = 0; i < Selection.Count(); i++)
		{
			Selection[i]->SetTargetPos(Target.Pos);
		}
}

}
